--[[
Name: Toolbox-1.0
Revision: $Rev: 24648 $
Author: Kodewulf (kodewulf@gmail.com)
Website: http://www.wowace.com
Description: Collection of various utility functions
Dependencies: AceLibrary, AceOO-2.0
]]

local MAJOR_VERSION = "Toolbox-1.0"
local MINOR_VERSION = "$Rev: 24648 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end
if not AceLibrary:HasInstance("AceOO-2.0") then error(MAJOR_VERSION .. " requires AceOO-2.0") end 
if not AceLibrary:HasInstance("AceLocale-2.2") then error(MAJOR_VERSION .. " requires AceLocale-2.2") end

local function print(text, r, g, b, frame, delay)
	(frame or DEFAULT_CHAT_FRAME):AddMessage(text, r, g, b, 1, delay or 5)
end

local function safecall(func,...)
	local success, err = pcall(func,...)
	if not success then geterrorhandler()(err:find("%.lua:%d+:") and err or (debugstack():match("\n(.-: )in.-\n") or "") .. err) end
end

local L = AceLibrary("AceLocale-2.2"):new(MAJOR_VERSION)

L:RegisterTranslations("enUS", function() return {
    ["zero"] = true,
    ["one"] = true,
    ["two"] = true,
    ["three"] = true,
    ["four"] = true,
    ["five"] = true,
    ["six"] = true,
    ["seven"] = true,
    ["eight"] = true,
    ["nine"] = true,
    ["ten"] = true,
    ["eleven"] = true,
    ["twelve"] = true,
    ["thirteen"] = true,
    ["fourteen"] = true,
    ["fifteen"] = true,
    ["sixteen"] = true,
    ["seventeen"] = true,
    ["eighteen"] = true,
    ["nineteen"] = true,
    ["twenty"] = true,
    ["thirty"] = true,
    ["forty"] = true,
    ["fifty"] = true,
    ["sixty"] = true,
    ["seventy"] = true,
    ["eighty"] = true,
    ["ninety"] = true,
    ["hundred"] = true,
    ["thousand"] = true,
    ["million"] = true,
    ["and"] = true,
    ["dot"] = true,
    ["minus"] = true,
    
    ["Alliance"] = true,
    ["Horde"] = true,
    ["of"] = true,
    ["Character"] = true,
    ["Realm"] = true,
    ["Class"] = true,
    ["Race"] = true,
    ["Faction"] = true,
    ["Default"] = true,
    
} end)

local AceOO = AceLibrary:GetInstance("AceOO-2.0")

--[[
local Toolbox = AceOO.Mixin {"Trim",
                       "RTrim",
                       "LTrim",
                       "Split",
                       "Reverse",
                       "Commify",
                       "Wordify",
                       "Clone",
                       "UnitRealm",
                       "UnitRace",
                       "UnitFaction",
                       "UnitClass",
                       "UnitName",
                       "UnitLevel",
                       "UnitCharacterID",
                       "UnitRealmID",
                       "UnitClassID",
                       "CustomDebug",
                       "Debug",
                       "IsDebugging",
                       "SetDebugging"
}

function Toolbox:CustomDebug(r, g, b, frame, delay, a1, ...)
    local tmp = {}
    
    if not self.debugging then
        return
    end

    local output = ("|cff7fff7f(DEBUG) %s:[%s.%3d]|r"):format(tostring(self), date("%H:%M:%S"), (GetTime() % 1) * 1000)
    
    a1 = tostring(a1)
    if a1:find("%%") and select('#', ...) >= 1 then
        for i = 1, select('#', ...) do
            tmp[i] = tostring((select(i, ...)))
        end
        output = output .. " " .. a1:format(unpack(tmp))
        for i = 1, select('#', ...) do
            tmp[i] = nil
        end
    else
        -- This block dynamically rebuilds the tmp array stopping on the first nil.
        tmp[1] = output
        tmp[2] = a1
        for i = 1, select('#', ...) do
            tmp[i+2] = tostring((select(i, ...)))
        end
        
        output = table.concat(tmp, " ")
        
        for i = 1, select('#', ...) + 2 do
            tmp[i] = nil
        end
    end

    print(output, r, g, b, frame or self.debugFrame, delay)
end

function Toolbox:Debug(...)
    Toolbox.CustomDebug(self, nil, nil, nil, nil, nil, ...)
end

function Toolbox:IsDebugging()
    return self.debugging
end

function Toolbox:SetDebugging(debugging)
    self.debugging = debugging

    if debugging then
        if type(self.OnDebugEnable) == "function" then
            safecall(self.OnDebugEnable, self)
        end
    else
        if type(self.OnDebugDisable) == "function" then
            safecall(self.OnDebugDisable, self)
        end
    end
end
--]]
local Toolbox = AceOO.Mixin {"Trim",
                       "RTrim",
                       "LTrim",
                       "Split",
                       "Reverse",
                       "Commify",
                       "Wordify",
                       "Clone",
                       "UnitRealm",
                       "UnitRace",
                       "UnitFaction",
                       "UnitClass",
                       "UnitName",
                       "UnitLevel",
                       "UnitCharacterID",
                       "UnitRealmID",
                       "UnitClassID"
}
                                                        
-- Trim whitespace from beginning and end of string
function Toolbox:Trim(s)
    return (string.gsub(s, "^%s*(.-)%s*$", "%1"))
end

-- Trim whitepace from beginning of string
function Toolbox:LTrim(s)
    return (string.gsub(s, "^%s*(.-)%s*", "%1"))
end

-- Trim whitespace from end of string
function Toolbox:RTrim(s)
    return (string.gsub(s, "%s*(.-)%s*$", "%1"))
end

-- Split a string based on pattern into a table
-- optionally trim each element
function Toolbox:Split(str, pat, trim)
    local ftrim = function (s) return s end
    if (not trim) then trim = true end
    if (trim == true) then ftrim = function (s) return self:Trim(s) end end

    local t = {}  -- NOTE: use {n = 0} in Lua-5.0
    local fpat = "(.-)" .. pat
    local last_end = 1
    local s, e, cap = string.find(str, fpat, 1)
    while s do
        if s ~= 1 or cap ~= "" then
            table.insert(t, ftrim(cap))
        end
        last_end = e+1
        s, e, cap = string.find(str, fpat, last_end)
    end
    if last_end <= string.len(str) then
        cap = string.sub(str, last_end)
        table.insert(t, ftrim(cap))
    end
    
    return t
end

-- Reverse a string
-- Reverse("teststring") will return "gnirtstset"
function Toolbox:Reverse(fs)
    return string.reverse(fs)
end

-- Adds commas to numbers
-- Commify(10000) returns "10,000" as string
function Toolbox:Commify(number)
    if (not number) then return "" end
    if (type(number) == "string") then return number end
    if (type(number) ~= "number") then return "" end

    if (string.len(number) < 1) then return "" end
    local ns = self:Split(tostring(number), "%.")

    local rs = self:Reverse(ns[1])
    local cs = ""
    local sc = 0
    
    for i = string.len(rs), 0, -1 do
        if (sc == 3) and (i > 0) then
            cs = cs .. ","
            sc = 0
        end
        
        cs = cs .. string.sub(ns[1], i, i)
        sc = sc + 1
    end
    
    cs = self:Reverse(cs)
    if (ns[2] ~= nil) then cs = cs .. "." .. ns[2] end
    return cs
end

-- Number to words
-- Wordify(10000) returns "ten thousand" as string
-- Decimals are supported
function Toolbox:Wordify(number)
    if (not number) then return "" end
    local nt = self:Split(tostring(number), "%.")
    
    local nwords = {}
    nwords[0] = L["zero"]
    nwords[1] = L["one"]
    nwords[2] = L["two"]
    nwords[3] = L["three"]
    nwords[4] = L["four"]
    nwords[5] = L["five"]
    nwords[6] = L["six"]
    nwords[7] = L["seven"]
    nwords[8] = L["eight"]
    nwords[9] = L["nine"]
    nwords[10] = L["ten"]
    nwords[11] = L["eleven"]
    nwords[12] = L["twelve"]
    nwords[13] = L["thirteen"]
    nwords[14] = L["fourteen"]
    nwords[15] = L["fifteen"]
    nwords[16] = L["sixteen"]
    nwords[17] = L["seventeen"]
    nwords[18] = L["eighteen"]
    nwords[19] = L["nineteen"]
    nwords[20] = L["twenty"]
    nwords[30] = L["thirty"]
    nwords[40] = L["forty"]
    nwords[50] = L["fifty"]
    nwords[60] = L["sixty"]
    nwords[70] = L["seventy"]
    nwords[80] = L["eighty"]
    nwords[90] = L["ninety"]
        
    local function intToWords(number)
        --[[
            Ported from PHP code
            --------------------
            
            lixlpixel PHParadise
            http://www.fundisom.com/phparadise/php/string_handling/number_to_words
        --]]
    
        if (type(number) ~= "number") then return "#" end
        if (math.fmod(number, 1) ~= 0) then return "#" end
        
        local words = ''
        if (number < 0) then
            words = L["minus"] .. ' '
            number = -number
        end
        
        if (number < 21) then 
            words = words .. nwords[number]
        elseif (number < 100) then
            words = words .. nwords[10 * math.floor(number / 10)]
            r = math.fmod(number, 10)
            if (r > 0) then words = words .. '-' .. nwords[r] end
        elseif (number < 1000) then
            words = words .. nwords[math.floor(number / 100)] .. ' ' .. L["hundred"]
            r = math.fmod(number, 100)
            if (r > 0) then words = words .. ' and ' .. intToWords(r) end
        elseif (number < 1000000) then
            words = words .. intToWords(math.floor(number / 1000)) .. ' ' .. L["thousand"]
            r = math.fmod(number, 1000)
            if (r > 0) then
                words = words .. ' '
                if (r < 100) then words = words .. L["and"] .. ' ' end
                words = words .. intToWords(r)
            end
        else
            words = words .. intToWords(math.floor(number / 1000000)) .. ' ' .. L["million"]
            r = math.fmod(number, 1000000)
            if (r > 0) then
                words = words .. ' '
                if (r < 100) then words = words .. L["and"] .. ' ' end
                words = words .. intToWords(r)
            end
        end

        return words
    end
    
    local ns = intToWords(tonumber(nt[1]))
    if (nt[2] ~= nil) then
        ns = ns .. ' ' .. L["dot"] 
        for i = 1, string.len(nt[2]) do
            ns = ns .. ' ' .. nwords[tonumber(string.sub(nt[2], i, i))]
        end
    end

    return ns
end

-- Clones a table
-- if 'deep' is false or omitted a shallow copy will be returned
function Toolbox:Clone(source, deep)
    if (not source) then return end
    if (type(source) ~= "table") then return source end
    
    if (not deep) then deep = false end
    if (type(deep) ~= "boolean") then deep = false end

    local function shallowCopy(source)
        local target = {}

        for k,v in pairs(source) do
            if (v ~= nil) then
                if (type(v) == "table") then
                    target[k] = shallowCopy(v)
                else
                    target[k] = v
                end
            end
        end
   
        return target
    end
    
    local function deepCopy(source, lookup)
        lookup = lookup or {}
        local target = {}
        
        if (not lookup[source]) then lookup[source] = source end
        for k, v in pairs(source) do
            if (type(k) == "table") then
                if (lookup[k]) then k = lookup[k]
                else k = deepCopy(k, lookup) end
            end
            if (type(v) ~= "table") then target[k] = v
            else
                if (lookup[v]) then target[k] = lookup[v]
                else target[k] = deepCopy(v, lookup) end
            end
        end
   
        return target
    end
    
    local target = {}
    if (not deep) then target = shallowCopy(source)
    else target = deepCopy(source) end
    
    return target
end

function Toolbox:UnitRace()
    local _,race = UnitRace("player")
    return race
end

function Toolbox:UnitFaction()
    local faction
    local race = Toolbox:UnitRace()
    
    if race == "Orc" or race == "Scourge" or race == "Troll" or race == "Tauren" or race == "BloodElf" then
        faction = L["Horde"]
    else
        faction = L["Alliance"]
    end
    
    return faction
end

function Toolbox:UnitRealm()
    return GetRealmName():trim()
end

function Toolbox:UnitClass()
    return UnitClass("player")
end

function Toolbox:UnitName()
    return UnitName("player")
end

function Toolbox:UnitLevel()
    return UnitLevel("player")
end

function Toolbox:UnitCharacterID()
    local server = Toolbox:UnitRealm()
    local charid = string.format("%s " .. L["of"] .. "%s", Toolbox:UnitName(), Toolbox:UnitRealm())
    return charid
end

function UnitRealmID()
    return Toolbox:UnitRealm() .. " - " .. Toolbox:UnitFaction()
end

function UnitClassID()
    return Toolbox:UnitClass()
end

AceLibrary:Register(Toolbox, MAJOR_VERSION, MINOR_VERSION, Toolbox.activate)
Toolbox = AceLibrary(MAJOR_VERSION)

