assert(School, "School not found!")

local _, _, REVISION = string.find("$Revision: 23349 $", "(%d+)")
local lang = AceLibrary("AceLocale-2.2"):new("SchoolFU")
local dewdrop = AceLibrary("Dewdrop-2.0")
local abacus = AceLibrary("Abacus-2.0")
local crayon = AceLibrary("Crayon-2.0")
local tablet = AceLibrary("Tablet-2.0")

lang:RegisterTranslations("enUS", function() return {
	["FuBar"] = true,

	["Show"] = true,
	["Show school data."] = true,
} end)

SchoolOptions.fullFubar = true

function SchoolOptions:OnTextUpdate()
    if (SchoolOptions.hasNoText) then return end
    
    self.schoolResults = self:FetchResults()
    if (not self.schoolResults) then return end
    if (not self.schoolResults.count) then self.schoolResults.count = 0 end
    if (not self.schoolResults.totalcost) then self.schoolResults.totalcost = 0 end
    
    
    local textCount
    local textTotalCost
    local textBar
    
    if (self.schoolResults.count > 0) then textCount = crayon:Green(self.schoolResults.count)
    else textCount = crayon:Red(self.schoolResults.count) end
    
    textTotalCost = abacus:FormatMoneyFull(self.schoolResults.totalcost)
    
    textBar = textCount
    if (self.schoolResults.count > 0) then textBar = textBar .. ": " .. textTotalCost end
    
    self:SetText(textBar)
end

function SchoolOptions:CreateUI()
    self:UpdateResults()
end


function SchoolOptions:FetchResults()
    -- School:Debug("SchoolOptions::FetchResults")
    School:GotoSchool()
    local school = School:Clone(School:FetchResults())
    if (not school) then return end
    
    results = {}
    results.data = {}
    
    for k,v in pairs(school.data) do
        if (v ~= nil) then
            if (not results.data[v.category]) then results.data[v.category] = {} end
            if (not results.data[v.category][k]) then results.data[v.category][k] = {} end
            results.data[v.category][k] = v
        end
    end
    
    results.totals = school.totals
    results.count = school.count
    results.totalcost = school.totalcost
    
    return results
end


function SchoolOptions:UpdateResults()
    -- School:Debug("SchoolOptions::UpdateResults")
    
    self.schoolResults = self:FetchResults()
    if (not self.schoolResults) then return end
    if (not self.schoolResults.count) then self.schoolResults.count = 0 end
    if (not self.schoolResults.totalcost) then self.schoolResults.totalcost = 0 end
    -- if (not SchoolOptions.hasNoText) then self:SetText(string.format("New skills: %d", self.schoolResults.count)) end
    if (not SchoolOptions.hasNoText) then
        local textCount
        local textTotalCost
        local textBar
        
        if (self.schoolResults.count > 0) then textCount = crayon:Green(self.schoolResults.count)
        else textCount = crayon:Red(self.schoolResults.count) end
        
        textTotalCost = abacus:FormatMoneyFull(self.schoolResults.totalcost)
        
        textBar = textCount
        if (self.schoolResults.count > 0) then textBar = textBar .. ": " .. textTotalCost end
        
        self:SetText(textBar)
    end
    
    -- if (not SchoolOptions.hasNoText) then self:SetText(string.format("%d: %s", crayon:Green(self.schoolResults.count), abacus:FormatMoneyFull(self.schoolResults.totalcost, true, true))) end
    if (self.schoolResults.count < 1) then return end
    if (not self.frame:IsShown()) then return end
    
    self.studentMoney = GetMoney()
    self.numResults = self.schoolResults.count
    self.totalCost = self.schoolResults.totalcost

    -- School:Debug("SchoolOptions::numResults ==> %s", self.numResults)
    
    for k,v in pairs(self.schoolResults.data) do
        if (v ~= nil) then
            -- School:Debug("Key ==> %s", k)
            if (type(v) == "table") then
                local cat = tablet:AddCategory(
                    'id', k,
                    'text', k,
                    'columns', 4,
                    'child_textR', 1,
                    'child_textG', 1,
                    'child_textB', 0,
                    'child_textR2', 1,
                    'child_textG2', 1,
                    'child_textB2', 1,
                    'child_justify3', "LEFT"
                )
                
                for vk, vv in pairs(v) do
                    if (vv ~= nil) and (type(v) == "table") then
                        -- School:Debug("Value Key ==> %s", vk)
                        
                        local result = School:Clone(vv)
                        result.formatted = {}
                        
                        if (not result.met) then result.formatted.fullname = crayon:Purple(result.fullname)
                        else result.formatted.fullname = crayon:Green(result.fullname) end
                        if (result.known) then result.formatted.fullname = crayon:Silver(result.fullname) end

                        result.formatted.cost = {}
                        if (GetMoney() < result.cost.money) then result.formatted.cost.money = crayon:Red(abacus:FormatMoneyFull(result.cost.money, false, false))
                        else result.formatted.cost.money = abacus:FormatMoneyFull(result.cost.money, true, true) end

                        if (result.trainers ~= nil) then
                            -- School:Debug("Found trainers.")
                            for tk, tv in pairs(result.trainers) do
                                if (tv ~= nil) then
                                    if (not result.formatted) then result.formatted = {} end
                                    if (not result.formatted.trainers) then result.formatted.trainers = {} end
                                    
                                    local trainer = {}
                                    trainer.name = crayon:Yellow(tv.name)
                                    trainer.type = tv.type
                                    if (tv.zone ~= nil) then 
                                        trainer.zone = tv.zone 
                                        trainer.location = tv.zone
                                    end
                                    if (tv.subzone ~= nil) then 
                                        trainer.subzone = tv.subzone
                                        trainer.location = trainer.subzone .. " (" .. trainer.location .. ")"
                                    end
                                    
                                    result.formatted.trainers[tk] = trainer
                                
                                    --[[
                                    if (not result.trainers) then result.trainers = {} end
                                    if (not result.trainers[vvk]) then result.trainers[vvk] = {} end

                                    if (not result.formatted.trainers) then result.formatted.trainers = {} end
                                    if (not result.formatted.trainers[vvk]) then result.formatted.trainers[vvk] = {} end
                                    
                                    result.trainers[vvk].name = vvv.name
                                    result.trainers[vvk].type = vvv.type
                                    result.trainers[vvk].zone = vvv.zone
                                    result.trainers[vvk].subzone = vvv.subzone
                                    
                                    result.formatted.trainers[vvk].name = crayon:Yellow(result.trainers[vvk].name)
                                    result.formatted.trainers[vvk].type = result.trainers[vvk].type
                                    if (result.trainers[vvk].zone ~= nil) then
                                        result.formatted.trainers[vvk].location = result.trainers[vvk].zone
                                    end
                                    if (result.trainers[vvk].subzone ~= nil) then
                                        result.formatted.trainers[vvk].location = result.trainers[vvk].subzone .. " (" .. result.formatted.trainers[vvk].location .. ")"
                                    end
                                    --]]
                                end
                            end
                        end

                        if (result ~= nil) then
                            cat:AddLine(
                                'text', result.formatted.fullname,
                                'text2', "     ",
                                'text3', "     ",
                                'text4', result.formatted.cost.money
                            )
                            
                            if (result.formatted.trainers ~= nil) then
                                for rtk, rtv in pairs(result.formatted.trainers) do
                                    if (rtv ~= nil) then
                                        cat:AddLine(
                                            'text', rtv.name,
                                            'text2', "     ",
                                            'text3', rtv.type,
                                            'text4', rtv.location
                                        )
                                    end
                                end
                            end
                        end
                        --]]
                    end
                end
            end
        end
    end

    -- if (not SchoolOptions.hasNoText) then self:SetText(string.format("New skills: %d", self.numResults)) end
end
