assert(School, "School not found!")

local _, _, REVISION = string.find("$Revision: 23349 $", "(%d+)")

SchoolLocale.notes = AceLibrary("AceLocale-2.2"):new("SchoolNotes")
SchoolLocale.notes:RegisterTranslations("enUS", function() return {
	["Notes"] = true,
	["Take notes"] = true,
    ["Toggles note taking."] = true,
    
    ["School"] = true,
    ["School Trainer"] = true,
    
    ["Reset"] = true,
    ["Reset SchoolNotes data"] = true,    
    
    ["Build"] = true,
    ["Build SchoolNotes data"] = true,
    
    ["Clear"] = true,
    ["Clear SchoolNotes data"] = true,
    
    ["Cartographer"] = true,
    ["Toggle Cartographer Notes"] = true,
    
    ["Metamap"] = true,
    ["Toggle Metamap Knowledgebase"] = true,
    
    ["MapNotes"] = true,
    ["Toggle MapNotes"] = true,
} end)

local lang = SchoolLocale.notes
SchoolNotes = School:NewModule(lang["Notes"])

SchoolNotes.defaults = {
	cartographer = false,
    metamap = false,
    mapnotes = false,
}

SchoolNotes.options = {
    cartographer = {
        type = "toggle",
        name = lang["Cartographer"],
        desc = lang["Toggle Cartographer Notes"],
        get = function() return SchoolNotes.db.profile.cartographer end,
        set = function(v) SchoolNotes.db.profile.cartographer = v end,
        hidden = function() return not Cartographer end,
        order = 120,
    },
    
    metamap = {
        type = "toggle",
        name = lang["Metamap"],
        desc = lang["Toggle Metamap Knowledgebase"],
        get = function() return SchoolNotes.db.profile.metamap end,
        set = function(v) SchoolNotes.db.profile.metamap = v end,
        hidden = function() return not Metamap end,
        order = 130,
    },
    
    mapnotes = {
        type = "toggle",
        name = lang["MapNotes"],
        desc = lang["Toggle MapNotes"],
        get = function() return SchoolNotes.db.profile.mapnotes end,
        set = function(v) SchoolNotes.db.profile.mapnotes = v end,
        hidden = function(v) return not MapNotes end,
        order = 140,
    },
    
    build = {
        type = "execute",
        name = lang["Build"],
        desc = lang["Build SchoolNotes data"],
        func = function() SchoolNotes:Build() end,
        disabled = function() return SchoolNotes.db.char.notes end,
        hidden = true,
        order = 200,
    },
    
    clear = {
        type = "execute",
        name = lang["Clear"],
        desc = lang["Clear SchoolNotes data"],
        func = function() SchoolNotes:Clear() end,
        disabled = function() return not SchoolNotes.db.char.notes end,
        order = 210,
    },
    
    reset = {
        type = "execute",
        name = lang["Reset"],
        desc = lang["Reset SchoolNotes data"],
        func = function() SchoolNotes:Reset() end,
        disabled = function() return not SchoolNotes.db.profile.notes end,
        order = 200,
    },
}

SchoolNotes:SetDebugging(School:IsDebugging())
SchoolNotes:SetDebugLevel(School:GetDebugLevel())

SchoolNotes.revision = tonumber(REVISION)

local function CartographerSetNote(note)
    if (Cartographer_Notes.SetNote ~= nil) and (type(Cartographer_Notes.SetNote) == "function") then
        -- note.creator = "School"
        Cartographer_Notes:SetNote(
            note.zone, note.x, note.y, note.icon, note.creator,
            'title', note.title,
            'titleR', note.titleR,
            'titleG', note.titleG,
            'titleB', note.titleB,
            'info', note.info,
            'infoR', note.infoR,
            'infoG', note.infoG,
            'infoB', note.infoB,
            'info2', note.info2,
            'info2R', note.info2R,
            'info2G', note.info2G,
            'info2B', note.info2B
        )
    end
end

local function CartographerGetNote(note)
    local result = {}
    
    if (Cartographer_Notes.GetNote ~= nil) and (type(Cartographer_Notes.GetNote) == "function") then
        result.zone, result.x, result.y, result.data, result.db = Cartographer_Notes:GetNote(note.zone, note.x, note.y)
    end
    
    return result
end

local function CartographerIsNote(note)
    local result = {}
    
    result = CartographerGetNote(note)
    
    if (result ~= nil) and (result.zone ~= nil) then
        return true
    end
    
    return false
end

local function CartographerDeleteNote(note)
    if (Cartographer_Notes.DeleteNote ~= nil) and (type(Cartographer_Notes.DeleteNote) == "function") then
        Cartographer_Notes:DeleteNote(note.zone, note.x, note.y)
    end
end

local function MetamapSetNote(note)
end

local function MetamapGetNote()
    local note = {}
    
    return note
end

local function MapnotesSetNote(note)
    if (not note) then return end

    --[[
    MapNotes:CreateNote(
        note.zone, note.x, note.y,
        note.title, "color",
        
        
    )
    --]]

    --[[
    MapNotes.CreateNote = function(zone, x, y, title, titleColor, category, iconID, description, descriptionColor, creator, creationDate)
	-- date: a 12 digit number as string as in ddmmyyHHMMSS
	-- lookup if the hash is already used, if it is - throw an error.
	local noteID = MapNotes.NoteHash(zone, x, y);
	if (MapNotesSaved_NoteDB[noteID]) then return false; end; -- THROW AN VISIBLE ERROR!! LOL
	if (not creationDate) then creationDate = date("%d%m%y%H%M%S"); end;
	if (not creator) then creator = UnitName("player").."-"..GetRealmName(); end;
	MapNotesSaved_NoteDB[noteID] = {
		zone = zone,
		x = x,
		y = y,
		title = title,
		description = description,
		category = category,
		iconID = iconID,
		misc = strjoin("$", titleColor, descriptionColor, creator, creationDate, date("%d%m%y%H%m%S")),
	};
    --]]
end
    
local function MapnotesGetNote()
    local note = {}
    
    return note
end

function SchoolNotes:OnInitialize()
    if (self.db.profile.cartographer == nil) then
        if (Cartographer) then self.db.profile.cartographer = true
        else self.db.profile.cartographer = false end
    end
    
    if (self.db.profile.metamap == nil) then
        if (Metamap) then self.db.profile.metamap = true
        else self.db.profile.metamap = false end
    end
    
    if (self.db.profile.mapnotes == nil) then
        if (MapNotes) then self.db.profile.mapnotes = true
        else self.db.profile.mapnotes = false end
    end

    if (not self.db.profile.notes) then self.db.profile.notes = {} end
end

function SchoolNotes:OnEnable()
    self.core.modules.notes = self

    if (self.db.profile.cartographer) and Cartographer_Notes then
        Cartographer_Notes:RegisterNotesDatabase(lang["School"], self.db.profile.notes, self)        
        Cartographer_Notes:RegisterIcon("SchoolTrainer", {
            text = lang["School Trainer"],
            path = "Interface\\AddOns\\School\\Icons\\School_Trainer",
        })
    end
    
    self:RegisterEvent("School_NotesCreate", "SetNote", 2)
end

function SchoolNotes:OnDisable()
    self:UnregisterAllEvents()
    
    if (self.db.profile.cartographer) and Cartographer_Notes then
        Cartographer_Notes:UnregisterNotesDatabase(lang["School"])
        Cartographer_Notes:UnregisterIcon("SchoolTrainer")
    end
end

function SchoolNotes:SetNote(note)
    self:Debug("SetNote")
    if (not note) then return end
    SetMapToCurrentZone()
    if (not note.zone) then note.zone = GetRealZoneText() end
    if (not note.x) or (not note.y) then note.x, note.y = GetPlayerMapPosition("player") end

    if (not note.id) then note.id = note.zone .. "#" .. note.x .. "#" .. note.y end

    if (not self.db.char.rollcall) then self.db.char.rollcall = {} end
    if (not self.db.char.rollcall[note.id]) then self.db.char.rollcall[note.id] = {} end
    if (self.db.char.rollcall[note.id].info ~= nil) then return end
    -- self:Debug("No existing note found.")
    
    if (not note.titleR) then note.titleR = 0 end
    if (not note.titleG) then note.titleG = 1 end
    if (not note.titleB) then note.titleB = 0 end
    
    if (not note.info) then note.info = "" end
    if (not note.infoR) then note.infoR = 1 end
    if (not note.infoG) then note.infoG = 1 end
    if (not note.infoB) then note.infoB = 1 end
    
    if (not note.info2) then note.info2 = "" end
    if (not note.info2R) then note.info2R = 1 end
    if (not note.info2G) then note.info2G = 1 end
    if (not note.info2B) then note.info2B = 1 end
    
    if (not note.creator) then note.creator = lang["School"] end
    if (not note.icon) then note.icon = "SchoolTrainer" end
    
    if (self.db.profile.cartographer) and (Cartographer_Notes) then CartographerSetNote(note) end
    
    if (self.db.profile.metamap) and (Metamap) then MetamapSetNote(note) end
    
    if (self.db.profile.mapnotes) and (MapNotes) then MapnotesSetNote(note) end
    
    self.db.char.rollcall[note.id].info = {note.zone, note.x, note.y}
end

function SchoolNotes:DeleteNote(note)
    if (self.db.profile.cartographer) and (Cartographer_Notes) then CartographerDeleteNote(note) end
end

function SchoolNotes:IsNote(note)
    local result = false
    
    if (self.db.profile.cartographer) and (Cartographer_Notes) then result = CartographerIsNote(note) end
    
    return result
end

function SchoolNotes:GetNote(note)
    if (self.db.profile.cartographer) and (Cartographer_Notes) then return CartographerGetNote(note) end
end

function SchoolNotes:Clear()
    if (not self.db.char.rollcall) then return end
    for k,v in pairs(self.db.char.rollcall) do
        if (v ~= nil) then
            local note = {}
            note.zone = v.info[1]
            note.x = v.info[2]
            note.y = v.info[3]
            
            if (self:IsNote(note)) then self:DeleteNote(note) end
        end
    end

    self.db.char.rollcall = nil
end

function SchoolNotes:Reset()
    if (self.db.char.rollcall) then
        if (not SchoolTeacher) then return end
        for k,v in pairs(self.db.char.rollcall) do
            if (v ~= nil) then
                if (SchoolTeacher.db.class.trainers ~= nil) and (SchoolTeacher.db.class.trainers[k] ~= nil) then
                    local trainer = SchoolTeacher.db.class.trainers[k]
                    local note = {}
                    note.id = trainer.name
                    note.title = trainer.name
                    note.info = trainer.type or ""
                    note.zone = v.info[1]
                    note.x = v.info[2]
                    note.y = v.info[3]
                    if (not self:IsNote(note)) then self:SetNote(note) end
                end
            end
        end
    end
end

function SchoolNotes:Purge()
    self.db.profile.notes = {}
    self.db.char.rollcall = {}
end

function SchoolNotes:Build()
end
