assert(School, "School not found!")

local _, _, REVISION = string.find("$Revision: 23349 $", "(%d+)")
local lang = AceLibrary("AceLocale-2.2"):new("SchoolTeacher")
local tooltip = AceLibrary("Gratuity-2.0")
-- local bt = AceLibrary("Babble-SpellTree-2.2")
-- local bs = AceLibrary("Babble-Spell-2.2")
-- local bz = AceLibrary("Babble-Zone-2.2")

lang:RegisterTranslations("enUS", function() return {
	["Teacher"] = true,

	["Rebuild"] = true,
	["Rebuild teacher data."] = true,
    
    ["Take Notes"] = true,
    ["Toggle note taking."] = true,
} end)

SchoolTeacher = School:NewModule(lang["Teacher"])
SchoolTeacher.notToggable = true

SchoolTeacher.defaults = {
    takeNotes = true,
}

SchoolTeacher.options = {
    takenotes = {
        type = "toggle",
        name = lang["Take Notes"],
        desc = lang["Toggle note taking."],
        get = function() return SchoolTeacher.db.profile.takeNotes end,
        set = function(v) SchoolTeacher.db.profile.takeNotes = v end,
        order = 100,
    },
}

SchoolTeacher:SetDebugging(School:IsDebugging())
SchoolTeacher:SetDebugLevel(School:GetDebugLevel())

SchoolTeacher.revision = tonumber(REVISION)

SchoolTeacher.npc = {}
SchoolTeacher.visiting = false

--[[
SchoolTeacher.merchant = {}
SchoolTeacher.merchant.itemsPerPage = MERCHANT_ITEMS_PER_PAGE or 10
SchoolTeacher.trainer = {}
--]]

function SchoolTeacher:OnInitialize()
end

function SchoolTeacher:OnEnable()
    self.core.modules.teacher = self
    self:RegisterEvents()
    self.class = self.core.classid
end

function SchoolTeacher:OnDisable()
    self.core.modules.teacher = nil
    self:UnregisterEvents()
end

function SchoolTeacher:RegisterEvents()
    if (not self.eventsLoaded) then
        self:RegisterEvent("TRAINER_SHOW", 2)
        self:RegisterEvent("TRAINER_CLOSED", 2)
        
        self:RegisterEvent("MERCHANT_SHOW", 2)
        self:RegisterEvent("MERCHANT_UPDATE")
        self:RegisterEvent("MERCHANT_CLOSED", 2)
        
        self:RegisterEvent("School_TeacherTrainerChanged", "Event_TeacherDataChanged", 2)
        self:RegisterEvent("School_TeacherMerchantChanged", "Event_TeacherDataChanged", 2)
        self:RegisterEvent("School_TeacherTakeNote", "Event_TeacherTakeNote", 5)
    end
    self.eventsLoaded = true
end

function SchoolTeacher:UnregisterEvents()
    if (self.eventsLoaded) then
        self:UnregisterAllEvents()
    end
    self.eventsLoaded = false
end

function SchoolTeacher:GetData()
    local data = {}

    data.charspells = self.db.char.spells
    data.spells = self.db.account.spells
    data.trainers = self.db.class.trainers
    
    data.counters = {}
    data.counters.spells = 0
    data.counters.trainers = 0
    data.counters.charspells = 0
    if (data.spells ~= nil) then
        for k,v in pairs(data.spells) do
            if (v ~= nil) then
                for vk, vv in pairs(v) do
                    if (vv ~= nil) then
                        data.counters.spells = data.counters.spells + 1
                    end
                end
            end
        end
    end
    if (data.trainers ~= nil) then
        for k,v in pairs(data.trainers) do
            if (v ~= nil) then
                data.counters.trainers = data.counters.trainers + 1
            end
        end    
    end
    if (data.charspells ~= nil) then
        for k,v in pairs(data.charspells) do
            if (v ~= nil) then
                for vk, vv in pairs(v) do
                    if (vv ~= nil) then
                        data.counters.charspells = data.counters.charspells + 1
                    end
                end
            end
        end
    end
    
    return data
end

function SchoolTeacher:TRAINER_SHOW()
    self:Debug("TRAINER_SHOW")
    self.visiting = true
    self.npc = self:ScanTeacherNPC()
end

function SchoolTeacher:TRAINER_UPDATE()
    self:Debug("TRAINER_UPDATE")
end

function SchoolTeacher:TRAINER_CLOSED()
    self:Debug("TRAINER_CLOSED")
    if (self.visiting) then self:ScanTrainer() end
    self.visiting = false
end

function SchoolTeacher:MERCHANT_SHOW()
    self:Debug("MERCHANT_SHOW")
    self.visiting = true
    self.npc = self:ScanTeacherNPC()
    self:ScanMerchant()
end

function SchoolTeacher:MERCHANT_UPDATE()
    self:Debug("MERCHANT_UPDATE")
    self:ScanMerchant()
end

function SchoolTeacher:MERCHANT_CLOSED()
    self:Debug("MERCHANT_CLOSED")
    if (self.visiting) then self:ScanMerchant() end
    self.visiting = false
end

function SchoolTeacher:Event_TeacherDataChanged()
    self:TriggerEvent("School_TeacherChanged")
end

function SchoolTeacher:Event_TeacherTakeNote(npc)
    if (self.db.profile.takeNotes) then self:NoteIt(npc) end
end

function SchoolTeacher:ScanTeacherNPC()
    local npc = {}
    npc.name = self:GetTeacherName()
    npc.zone = self:GetTeacherZone()
    npc.subzone = self:GetTeacherSubZone()
    npc.type = self:GetTeacherType()
    npc.isteacher = self:IsTeacherNPC(npc)
    
    return npc
end

function SchoolTeacher:IsTeacherNPC(npc)
    local result = true
    
    local npcTypes = {}
    npcTypes["Demon Trainer"] = true
    
    local npcNames = {}
    npcNames["Daestra"] = true
    
    if (not npcTypes[npc.type]) then result = false end
    if (not result) and (npcNames[npc.name]) then result = true end
    
    return result
end

function SchoolTeacher:ScanTrainer()
    self:Debug("ScanTrainer")
    local spells = {}
    local trainers = {}
    local charspells = {}

    if (self.db.account.spells ~= nil) then spells = self.db.account.spells end
    if (self.db.class.trainers ~= nil) then trainers = self.db.class.trainers end
    if (self.db.char.spells ~= nil) then charspells = self:Clone(self.db.char.spells) end
    
    if (not self.npc) then self:ScanTeacherNPC() end
    
    local npc = self.npc

    local filter = {}
    filter.available = GetTrainerServiceTypeFilter("available")
    filter.unavailable = GetTrainerServiceTypeFilter("unavailable")
    filter.used = GetTrainerServiceTypeFilter("used")

    SetTrainerServiceTypeFilter("available", 1)
    SetTrainerServiceTypeFilter("unavailable", 1)
    SetTrainerServiceTypeFilter("used", 1)

    ExpandTrainerSkillLine(0)

    local numTrainerServices = GetNumTrainerServices()
    if (not numTrainerServices or (numTrainerServices < 1)) then return end
    
    self:Debug("Found %s trainer services.", numTrainerServices)

    local category = nil
    for i = 1, numTrainerServices do
        local continue = true
        local service = {}
        local charservice = {}

        service.name, service.subtext, charservice.type, service.expanded = GetTrainerServiceInfo(i)
        if (not service.name or not charservice.type) then continue = false end

        if (charservice.type == "header") then 
            if (bt) then service.name = bt[service.name] end
            category = service.name
            continue = false
        end

        if (continue) then 
            if (category ~= nil) then service.category = category end
            if (not service.subtext) or (string.len(service.subtext) < 1) then service.subtext = "None" end

            -- self:Debug("service.category ==> %s", service.category)
            -- self:Debug("service.name ==> %s", service.name)
            -- self:Debug("service.subtext ==> %s", service.subtext)
            
            if (bs) then service.name = bs[service.name] end

            if (not spells[service.category]) then spells[service.category] = {} end
            if (not spells[service.category][service.name]) then spells[service.category][service.name] = {} end
            if (not spells[service.category][service.name][service.subtext]) then spells[service.category][service.name][service.subtext] = {} end

            if (not charspells[service.category]) then charspells[service.category] = {} end
            if (not charspells[service.category][service.name]) then charspells[service.category][service.name] = {} end
            if (not charspells[service.category][service.name][service.subtext]) then charspells[service.category][service.name][service.subtext] = {} end

            if (not spells[service.category][service.name][service.subtext].fullname) then
                service.fullname = service.name
            end

            if (not spells[service.category][service.name][service.subtext].rank) then
                service.rank = 0

                if (service.subtext ~= "None") then
                    local fa, fx, rank = string.find(service.subtext, "Rank (%d+)")
                    service.rank = tonumber(rank) or 0
                    service.fullname = service.fullname .. " (" .. service.subtext .. ")"
                end
            else
                service.rank = spells[service.category][service.name][service.subtext].rank
                service.fullname = spells[service.category][service.name][service.subtext].fullname
            end

            if (not spells[service.category][service.name][service.subtext].icon) then
                service.icon = GetTrainerServiceIcon(i)
            else
                service.icon = spells[service.category][service.name][service.subtext].icon
            end

            if (not spells[service.category][service.name][service.subtext].level) then
                service.level = GetTrainerServiceLevelReq(i)
            else
                service.level = spells[service.category][service.name][service.subtext].level
            end

            if (not spells[service.category][service.name][service.subtext].req) then
                local req = {}
                req.skill, req.rank, req.hasReq = GetTrainerServiceSkillReq(i)
                if (req.skill) then service.req = req end
            else
                service.req = spells[service.category][service.name][service.subtext].req
            end

            if (not spells[service.category][service.name][service.subtext].ability) then
                local numRequirements = GetTrainerServiceNumAbilityReq(i);
                if (numRequirements > 0) then
                    service.ability = {}
                    for k = 1, numRequirements do
                        service.ability[k] = {}
                        service.ability[k].skill, service.ability[k].req = GetTrainerServiceAbilityReq(i, k)
                    end
                end
            else
                service.ability = spells[service.category][service.name][service.subtext].ability
            end

            if (not spells[service.category][service.name][service.subtext].desciption) then
                -- Fix wrapped descriptions
                service.description = GetTrainerServiceDescription(i)
            else
                service.description = spells[service.category][service.name][service.subtext].description
            end

            if (not spells[service.category][service.name][service.subtext].learnspell) or (not spells[service.category][service.name][service.subtext].petlearnspell) then
                service.learnspell, service.petlearnspell = IsTrainerServiceLearnSpell(i)
            else
                service.learnspell = spells[service.category][service.name][service.subtext].learnspell
                service.petlearnspell = spells[service.category][service.name][service.subtext].petlearnspell                
            end

            if (not spells[service.category][service.name][service.subtext].known) then
                if (charservice.type == "used") then charservice.known = true
                elseif (charservice.type == "available") or (charservice.type == "unavailable") then 
                    charservice.known = false
                end
            else
                charservice.known = spells[service.category][service.name][service.subtext].known
            end

            if (not service.cost) then service.cost = {} end
            -- Money, Character Point Cost 1, Character Point Cost 2
            service.cost.money, service.cost.cost1, service.cost.cost2 = GetTrainerServiceCost(i)

            spells[service.category][service.name][service.subtext] = service

            if (not trainers[npc.name]) then trainers[npc.name] = {} end
            if (not trainers[npc.name].coords) then
                SetMapToCurrentZone()
                npc.coords = {}
                npc.coords.x, npc.coords.y = GetPlayerMapPosition("player")
                -- self:Debug("Coords\nx: %s\ny: %s", npc.coords.x, npc.coords.y)
            end
            trainers[npc.name] = npc
            if (not trainers[npc.name].spells) then trainers[npc.name].spells = {} end
            if (not trainers[npc.name].spells[service.name]) then trainers[npc.name].spells[service.name] = {} end
            if (not trainers[npc.name].spells[service.name].rank) then trainers[npc.name].spells[service.name].rank = service.rank
            else
                if (service.rank > trainers[npc.name].spells[service.name].rank) then
                    trainers[npc.name].spells[service.name].rank = service.rank
                end
            end

            charspells[service.category][service.name][service.subtext].known = charservice.known

            -- self:Debug("\nservice.category ==> %s\nservice.name ==> %s\nservice.subtext ==> %s\nnpc.name ==> %s", service.category, service.name, service.subtext, npc.name)
        end
    end

    self.db.account.spells = spells
    self.db.class.trainers = trainers
    self.db.char.spells = charspells

    self:TriggerEvent("School_TeacherTrainerChanged")
    self:TriggerEvent("School_TeacherTakeNote", npc)
end

function SchoolTeacher:ScanMerchant()
    self:Debug("ScanMerchant")
    local spells = {}
    local trainers = {}
    local charspells = {}

    if (self.db.account.spells ~= nil) then spells = self.db.account.spells end
    if (self.db.class.trainers ~= nil) then trainers = self.db.class.trainers end
    if (self.db.char.spells ~= nil) then charspells = self:Clone(self.db.char.spells) end

    if (not self.npc) then self:ScanTeacherNPC() end
    
    local npc = self.npc
    self:Debug("NPC\nname: %s\ntype: %s\nisteacher: %s", npc.name, npc.type, npc.isteacher)

    if (npc.isteacher) then
        local category = "Pet"
        
        self:Debug("%s has %s items.", npc.name, GetMerchantNumItems())
        for i = 1, GetMerchantNumItems() do
            local continue = true
            local service = {}
            local charservice = {}
            
            service.cost = {}
            service.fullname, service.icon, service.cost.money, service.quantity, _,_ = GetMerchantItemInfo(i)
            
            if (not service.fullname) then continue = false end
            
            if (continue) then
                if (category ~= nil) then service.category = category end
                local fa, fx, subtext = string.find(service.fullname, "(Rank %d+)")
                service.subtext = subtext or "None"
                
                if (not fa) then service.name = service.fullname 
                else service.name = string.sub(service.fullname, 1, (fa - 2)) end
                
                if (bs) then service.name = bs[service.name] end

                -- ????
                -- local fa, fx = string.find(service.fullname, service.subtext)
                
                if (not spells[service.category]) then spells[service.category] = {} end
                if (not spells[service.category][service.name]) then spells[service.category][service.name] = {} end
                if (not spells[service.category][service.name][service.subtext]) then spells[service.category][service.name][service.subtext] = {} end

                if (not charspells[service.category]) then charspells[service.category] = {} end
                if (not charspells[service.category][service.name]) then charspells[service.category][service.name] = {} end
                if (not charspells[service.category][service.name][service.subtext]) then charspells[service.category][service.name][service.subtext] = {} end
                
                tooltip:SetMerchantItem(i)
                if (not charspells[service.category][service.name][service.subtext].known) then
                    local _,_, match = tooltip:Find("(Already known)", 2, 20)
                    if (match ~= nil) then charservice.known = true 
                    else charservice.known = false end
                else
                    charservice.known = charspells[service.category][service.name][service.subtext].known
                end
                
                if (not spells[service.category][service.name][service.subtext].level) then
                    local _,_, level = tooltip:Find("Requires Level (%d+)", 2, 20)
                    if (level ~= nil) then service.level = tonumber(level) end
                else
                    service.level = spells[service.category][service.name][service.subtext].level
                end
                
                if (not spells[service.category][service.name][service.subtext].rank) then
                    service.rank = 0
                    if (service.subtext ~= "None") then
                        local fa, fx, rank = string.find(service.subtext, "Rank (%d+)")
                        service.rank = tonumber(rank) or 0
                    end
                else
                    service.rank = spells[service.category][service.name][service.subtext].rank
                end

                if (not spells[service.category][service.name][service.subtext].icon) then
                else
                    service.icon = spells[service.category][service.name][service.subtext].icon
                end
               
                --[[
                self:Debug("spell.fullname ==> %s", service.fullname)
                self:Debug("service.name ==> %s", service.name)
                self:Debug("service.subtext ==> %s", service.subtext)
                self:Debug("service.rank ==> %s", service.rank)
                self:Debug("service.level ==> %s", service.level)
                --]]
                
                spells[service.category][service.name][service.subtext] = service                

                if (not trainers[npc.name]) then trainers[npc.name] = {} end
                if (not trainers[npc.name].coords) then
                    SetMapToCurrentZone()
                    npc.coords = {}
                    npc.coords.x, npc.coords.y = GetPlayerMapPosition("player")
                end
                trainers[npc.name] = npc
                if (not trainers[npc.name].spells) then trainers[npc.name].spells = {} end
                if (not trainers[npc.name].spells[service.name]) then trainers[npc.name].spells[service.name] = {} end
                if (not trainers[npc.name].spells[service.name].rank) then trainers[npc.name].spells[service.name].rank = service.rank 
                else
                    if (service.rank > trainers[npc.name].spells[service.name].rank) then
                        trainers[npc.name].spells[service.name].rank = service.rank
                    end
                end

                charspells[service.category][service.name][service.subtext].known = charservice.known 
            end
        end

        self.db.account.spells = spells
        self.db.class.trainers = trainers
        self.db.char.spells = charspells

        self:TriggerEvent("School_TeacherMerchantChanged")
        self:TriggerEvent("School_TeacherTakeNote", npc)
    end
end


function SchoolTeacher:LogTrainerInfo()
    -- if (self.IsAtTrainer) then
        self:LogClassTrainerInfo()
        self:TriggerEvent("School_Teacher_Changed")
    -- end
end

function SchoolTeacher:LogMerchantInfo()
    if (self.IsAtMerchant) then
        self:LogClassMerchantInfo()
        self:TriggerEvent("School_Teacher_Changed")
    end
end

function SchoolTeacher:LogTradeskillTrainerInfo()
    self:LevelDebug(3, "LogTradeskillTrainerInfo")
    
    self:TriggerEvent("School_Teacher_Changed")
end

function SchoolTeacher:LogClassTrainerInfo()
    self:Debug("LogClassTrainerInfo")
    local spells = {}
    local trainers = {}
    local charspells = {}

    if (self.db.account.spells ~= nil) then spells = self.db.account.spells end
    if (self.db.class.trainers ~= nil) then trainers = self.db.class.trainers end
    if (self.db.char.spells ~= nil) then charspells = self:Clone(self.db.char.spells) end
    
    local npc = {}
    npc.name = self:GetTeacherName()
    npc.zone = self:GetTeacherZone()
    npc.subzone = self:GetTeacherSubZone()
    npc.type = self:GetTeacherType()

    local filter = {}
    filter.available = GetTrainerServiceTypeFilter("available")
    filter.unavailable = GetTrainerServiceTypeFilter("unavailable")
    filter.used = GetTrainerServiceTypeFilter("used")

    SetTrainerServiceTypeFilter("available", 1)
    SetTrainerServiceTypeFilter("unavailable", 1)
    SetTrainerServiceTypeFilter("used", 1)

    ExpandTrainerSkillLine(0)

    local numTrainerServices = GetNumTrainerServices()
    if (not numTrainerServices or (numTrainerServices < 1)) then return end
    
    self:Debug("Found %s trainer services.", numTrainerServices)

    local category = nil
    for i = 1, numTrainerServices do
        local continue = true
        local service = {}
        local charservice = {}

        service.name, service.subtext, charservice.type, service.expanded = GetTrainerServiceInfo(i)
        if (not service.name or not charservice.type) then continue = false end

        if (charservice.type == "header") then 
            if (bt) then service.name = bt[service.name] end
            category = service.name
            continue = false
        end

        if (continue) then 
            if (category ~= nil) then service.category = category end
            if (not service.subtext) or (string.len(service.subtext) < 1) then service.subtext = "None" end

            -- self:Debug("service.category ==> %s", service.category)
            -- self:Debug("service.name ==> %s", service.name)
            -- self:Debug("service.subtext ==> %s", service.subtext)
            
            if (bs) then service.name = bs[service.name] end

            if (not spells[service.category]) then spells[service.category] = {} end
            if (not spells[service.category][service.name]) then spells[service.category][service.name] = {} end
            if (not spells[service.category][service.name][service.subtext]) then spells[service.category][service.name][service.subtext] = {} end

            if (not charspells[service.category]) then charspells[service.category] = {} end
            if (not charspells[service.category][service.name]) then charspells[service.category][service.name] = {} end
            if (not charspells[service.category][service.name][service.subtext]) then charspells[service.category][service.name][service.subtext] = {} end

            if (not spells[service.category][service.name][service.subtext].fullname) then
                service.fullname = service.name
            end

            if (not spells[service.category][service.name][service.subtext].rank) then
                service.rank = 0

                if (service.subtext ~= "None") then
                    local fa, fx, rank = string.find(service.subtext, "Rank (%d+)")
                    service.rank = tonumber(rank) or 0
                    service.fullname = service.fullname .. " (" .. service.subtext .. ")"
                end
            else
                service.rank = spells[service.category][service.name][service.subtext].rank
                service.fullname = spells[service.category][service.name][service.subtext].fullname
            end

            if (not spells[service.category][service.name][service.subtext].icon) then
                service.icon = GetTrainerServiceIcon(i)
            else
                service.icon = spells[service.category][service.name][service.subtext].icon
            end

            if (not spells[service.category][service.name][service.subtext].level) then
                service.level = GetTrainerServiceLevelReq(i)
            else
                service.level = spells[service.category][service.name][service.subtext].level
            end

            if (not spells[service.category][service.name][service.subtext].req) then
                local req = {}
                req.skill, req.rank, req.hasReq = GetTrainerServiceSkillReq(i)
                if (req.skill) then service.req = req end
            else
                service.req = spells[service.category][service.name][service.subtext].req
            end

            if (not spells[service.category][service.name][service.subtext].ability) then
                local numRequirements = GetTrainerServiceNumAbilityReq(i);
                if (numRequirements > 0) then
                    service.ability = {}
                    for k = 1, numRequirements do
                        service.ability[k] = {}
                        service.ability[k].skill, service.ability[k].req = GetTrainerServiceAbilityReq(i, k)
                    end
                end
            else
                service.ability = spells[service.category][service.name][service.subtext].ability
            end

            if (not spells[service.category][service.name][service.subtext].desciption) then
                -- Fix wrapped descriptions
                -- service.description = GetTrainerServiceDescription(i)
            else
                service.description = spells[service.category][service.name][service.subtext].description
            end

            if (not spells[service.category][service.name][service.subtext].learnspell) or (not spells[service.category][service.name][service.subtext].petlearnspell) then
                service.learnspell, service.petlearnspell = IsTrainerServiceLearnSpell(i)
            else
                service.learnspell = spells[service.category][service.name][service.subtext].learnspell
                service.petlearnspell = spells[service.category][service.name][service.subtext].petlearnspell                
            end

            if (not spells[service.category][service.name][service.subtext].known) then
                if (charservice.type == "used") then charservice.known = true
                elseif (charservice.type == "available") or (charservice.type == "unavailable") then 
                    charservice.known = false
                end
            else
                charservice.known = spells[service.category][service.name][service.subtext].known
            end

            if (not service.cost) then service.cost = {} end
            -- Money, Character Point Cost 1, Character Point Cost 2
            service.cost.money, service.cost.cost1, service.cost.cost2 = GetTrainerServiceCost(i)

            spells[service.category][service.name][service.subtext] = service

            if (not trainers[npc.name]) then trainers[npc.name] = {} end
            if (not trainers[npc.name].coords) then
                SetMapToCurrentZone()
                npc.coords = {}
                npc.coords.x, npc.coords.y = GetPlayerMapPosition("player")
                -- self:Debug("Coords\nx: %s\ny: %s", npc.coords.x, npc.coords.y)
            end
            trainers[npc.name] = npc
            if (not trainers[npc.name].spells) then trainers[npc.name].spells = {} end
            if (not trainers[npc.name].spells[service.name]) then trainers[npc.name].spells[service.name] = {} end
            if (not trainers[npc.name].spells[service.name].rank) then trainers[npc.name].spells[service.name].rank = service.rank
            else
                if (service.rank > trainers[npc.name].spells[service.name].rank) then
                    trainers[npc.name].spells[service.name].rank = service.rank
                end
            end

            charspells[service.category][service.name][service.subtext].known = charservice.known

            -- self:Debug("\nservice.category ==> %s\nservice.name ==> %s\nservice.subtext ==> %s\nnpc.name ==> %s", service.category, service.name, service.subtext, npc.name)
        end
    end

    self.db.account.spells = spells
    self.db.class.trainers = trainers
    self.db.char.spells = charspells
    
    if SchoolNotes and self.db.profile.takeNotes then self:NoteIt(npc) end
end

function SchoolTeacher:LogClassMerchantInfo()
    self:Debug("LogClassMerchantInfo")
    local spells = {}
    local trainers = {}
    local charspells = {}

    if (self.db.account.spells ~= nil) then spells = self.db.account.spells end
    if (self.db.class.trainers ~= nil) then trainers = self.db.class.trainers end
    if (self.db.char.spells ~= nil) then charspells = self:Clone(self.db.char.spells) end

    local npc = {}
    npc.name = self:GetTeacherName()
    npc.zone = self:GetTeacherZone()
    npc.subzone = self:GetTeacherSubZone()
    npc.type = self:GetTeacherType()

    if (self:NPCIsTeacher(npc)) then
        local category = "Pet"
        
        for i = 1, GetMerchantNumItems() do
            local continue = true
            local service = {}
            local charservice = {}
            
            service.cost = {}
            service.fullname, service.icon, service.cost.money, service.quantity, _,_ = GetMerchantItemInfo(i)
            
            if (not service.fullname) then continue = false end
            
            if (continue) then
                if (category ~= nil) then service.category = category end
                local fa, fx, subtext = string.find(service.fullname, "(Rank %d+)")
                service.subtext = subtext or "None"
                
                if (not fa) then service.name = service.fullname 
                else service.name = string.sub(service.fullname, 1, (fa - 2)) end
                
                if (bs) then service.name = bs[service.name] end

                -- ????
                -- local fa, fx = string.find(service.fullname, service.subtext)
                
                if (not spells[service.category]) then spells[service.category] = {} end
                if (not spells[service.category][service.name]) then spells[service.category][service.name] = {} end
                if (not spells[service.category][service.name][service.subtext]) then spells[service.category][service.name][service.subtext] = {} end

                if (not charspells[service.category]) then charspells[service.category] = {} end
                if (not charspells[service.category][service.name]) then charspells[service.category][service.name] = {} end
                if (not charspells[service.category][service.name][service.subtext]) then charspells[service.category][service.name][service.subtext] = {} end
                
                tooltip:SetMerchantItem(i)
                if (not charspells[service.category][service.name][service.subtext].known) then
                    local _,_, match = tooltip:Find("(Already known)", 2, 20)
                    if (match ~= nil) then charservice.known = true 
                    else charservice.known = false end
                else
                    charservice.known = charspells[service.category][service.name][service.subtext].known
                end
                
                if (not spells[service.category][service.name][service.subtext].level) then
                    local _,_, level = tooltip:Find("Requires Level (%d+)", 2, 20)
                    if (level ~= nil) then service.level = tonumber(level) end
                else
                    service.level = spells[service.category][service.name][service.subtext].level
                end
                
                if (not spells[service.category][service.name][service.subtext].rank) then
                    service.rank = 0
                    if (service.subtext ~= "None") then
                        local fa, fx, rank = string.find(service.subtext, "Rank (%d+)")
                        service.rank = tonumber(rank) or 0
                    end
                else
                    service.rank = spells[service.category][service.name][service.subtext].rank
                end

                if (not spells[service.category][service.name][service.subtext].icon) then
                else
                    service.icon = spells[service.category][service.name][service.subtext].icon
                end
               
                --[[
                self:Debug("spell.fullname ==> %s", service.fullname)
                self:Debug("service.name ==> %s", service.name)
                self:Debug("service.subtext ==> %s", service.subtext)
                self:Debug("service.rank ==> %s", service.rank)
                self:Debug("service.level ==> %s", service.level)
                --]]
                
                spells[service.category][service.name][service.subtext] = service                

                if (not trainers[npc.name]) then trainers[npc.name] = {} end
                if (not trainers[npc.name].coords) then
                    self:Debug("Checkpoint 1")
                    SetMapToCurrentZone()
                    self:Debug("Checkpoint 2")
                    npc.coords = {}
                    self:Debug("Checkpoint 3")
                    npc.coords.x, npc.coords.y = GetPlayerMapPosition("player")
                    self:Debug("Checkpoint 4")
                    self:Debug("Coords\nx: %s\ny: %s", npc.coords.x, npc.coords.y)
                end
                trainers[npc.name] = npc
                if (not trainers[npc.name].spells) then trainers[npc.name].spells = {} end
                if (not trainers[npc.name].spells[service.name]) then trainers[npc.name].spells[service.name] = {} end
                if (not trainers[npc.name].spells[service.name].rank) then trainers[npc.name].spells[service.name].rank = service.rank 
                else
                    if (service.rank > trainers[npc.name].spells[service.name].rank) then
                        trainers[npc.name].spells[service.name].rank = service.rank
                    end
                end

                charspells[service.category][service.name][service.subtext].known = charservice.known 
            end
        end

        self.db.account.spells = spells
        self.db.class.trainers = trainers
        self.db.char.spells = charspells

        if SchoolNotes and self.db.profile.takeNotes then self:NoteIt(npc) end
    end
end


function SchoolTeacher:GetTeacherName()
    return UnitName("npc")
end

function SchoolTeacher:GetTeacherZone(teacher)
    local zone = GetZoneText()
    if (bz) then zone = bz[zone] end
    return zone
end

function SchoolTeacher:GetTeacherSubZone(teacher)
    local subzone = GetSubZoneText()
    if (bz) and (string.len(subzone) > 0) then subzone = bz[subzone] end
    return subzone
end

function SchoolTeacher:GetTeacherType(teacher)
    local result 
    
    local tooltip = AceLibrary("Gratuity-2.0")
        
    tooltip:SetUnit("npc")
    result = tooltip:GetLine(2)
    
    if (not result) then
        if (not teacher) then teacher = self:GetTeacherName() end
    
        local teacherTypes = {}
        teacherTypes["Aldrae"] = "Priest Trainer"
        teacherTypes["Lotheolan"] = "Priest Trainer"
        teacherTypes["Belestra"] = "Priest Trainer"
        teacherTypes["Ponaris"] = "Priest Trainer"
        teacherTypes["Daestra"] = "Demon Trainer"
        
        result = teacherTypes[teacher]
    end

    -- self:Debug("GetTeacherType:result ==> %s", result)
    return result
end    

function SchoolTeacher:NPCIsTeacher(npc)
    local result = true
    
    local teacherTypes = {}
    teacherTypes["Demon Trainer"] = true
    
    
    local teacherNames = {}
    teacherNames["Daestra"] = true
    

    if (not teacherTypes[npc.type]) then result = false end
    if (not result) and (teacherNames[npc.name]) then result = true end
    
    return result
end

function SchoolTeacher:NoteIt(trainer)
    self:Debug("NoteIt")
    if (not trainer) then return end
    if (not SchoolNotes) then return end
    
    local note = {}
    note.id = trainer.name
    note.title = trainer.name
    note.info = trainer.type or ""
    note.creator = SchoolLocale.school["School"]
    -- note.icon = "Interface\\School\\Icons\\School_Trainer"
    if (trainer.coords) then
        if (trainer.coords.x) and (trainer.coords.y) then
            note.x = trainer.coords.x
            note.y = trainer.coords.y
        end
    end
    
    self:TriggerEvent("School_NotesCreate", note)
end

function SchoolTeacher:Upgrade()
    if (not self.db.account.spells) and (not self.db.class.trainers) and (not self.db.char.spells) then return true end
    
    function UpgradeVersion1()
        return true
    end
    
    function UpgradeVersion2()
        local continue = true
        local result = false
        if (not UpgradeVersion1) then continue = false end

        if (continue) then
            -- Add Upgrade Code here
        
        
        end
        
        return result
    end

    return UpgradeVersion2()
end








