assert(School, "School not found!")

local _, _, REVISION = string.find("$Revision: 23349 $", "(%d+)")
local lang = AceLibrary("AceLocale-2.2"):new("SchoolUI")
local dewdrop = AceLibrary("Dewdrop-2.0")
local abacus = AceLibrary("Abacus-2.0")
local crayon = AceLibrary("Crayon-2.0")
-- local deformat = AceLibrary("Deformat-2.0")
-- local tablet = AceLibrary("Tablet-2.0")

lang:RegisterTranslations("enUS", function() return {
    ["UI"] = true,
    ["User Interface"] = true,

    ["Show"] = true,
    ["Show school frame"] = true,

    ["Hide"] = true,
    ["Hide school frame"] = true,
    
    ["Close window"] = true,
    
    ["Lock position"] = true,
    ["Toggles locking the School frame in place."] = true,
    
    ["Fixed Tooltip"] = true,
    ["Locks the tooltip in place on the left side of the frame."] = true,
    
    ["Tooltip Position"] = true,
    ["Sets tooltip placement on left or right of frame."] = true,

    ["Detail Position"] = true,
    ["Sets detail placement on left or right of frame."] = true,

    ["Rows"] = true,
    ["Sets result row count.\n|cffff0000REQUIRES RELOADUI|r"] = true,
    
    ["Backdrop Alpha"] = true,
    ["Sets frame background transparency level."] = true,

    ["Ignore cost"] = true,
    ["Ignore learning cost."] = true,
    ["Toggle ignore learning cost."] = true,
    
    ["Ignore combat"] = true,
    ["Ignore combat."] = true,
    ["Toggle ignore combat."] = true,
    
    ["You can't afford this."] = true,
    
    ["Your level is too low."] = true,
    
    ["Cost"] = true,
    ["Category"] = true,
    ["Trainers"] = true,
    
    ["at"] = true,
    ["in"] = true,
    
    ["Requires level %d"] = true,
    ["Requires %s"] = true,
    
    ["Show frame"] = true,
    ["Show School frame"] = true,

    ["Talents"] = true,
    ["Combat"] = true,
    ["Login"] = true,
    
    ["Already known"] = true,
    
    ["Unknown"] = true,
    
} end)

SchoolUI = School:NewModule(lang["User Interface"])

SchoolUI.defaults = {
    locked = false,
    fixedTooltip = true,
    resultRows = 10,
    backgroundAlpha = 0.75,
    tooltipPosition = "left",
    detailPosition = "left",
}

SchoolUI.commands = {
    show = {
        type = 'toggle',
        name = lang["Show school frame"],
        desc = lang["Show school frame"],
        get = function() return SchoolUI:IsShown() end,
        set = function() 
            if (not SchoolUI:IsShown()) then
                SchoolUI:ShowUI()
            else
                SchoolUI:HideUI()
            end
        end,
        order = 99,
    }
}

SchoolUI.options = {
    lock = {
        type = "toggle",
        name = lang["Lock position"],
        desc = lang["Toggles locking the School frame in place."],
        get = function() return SchoolUI.db.profile.locked end,
        set = function(v) SchoolUI.db.profile.locked = v end,
        order = 120,
    },
    fixedtooltip = {
        type = "toggle",
        name = lang["Fixed Tooltip"],
        desc = lang["Locks the tooltip in place on the left side of the frame."],
        get = function() return SchoolUI.db.profile.fixedTooltip end,
        set = function(v) SchoolUI.db.profile.fixedTooltip = v end,
        order = 130,
    },
    tooltipposition = {
       type = "text",
       name = lang["Tooltip Position"],
       desc = lang["Sets tooltip placement on left or right of frame."],
       get = function() return SchoolUI.db.profile.tooltipPosition end,
       set = function(v) SchoolUI.db.profile.tooltipPosition = v end,
       validate = {
           ["left"] = "Left",
           ["right"] = "Right",
       },
       order = 140,
    },
    detailposition = {
        type = "text",
        name = lang["Detail Position"],
        desc = lang["Sets detail placement on left or right of frame."],
        get = function() return SchoolUI.db.profile.detailPosition end,
        set = function(v) SchoolUI.db.profile.detailPosition = v end,
        validate = {
            ["left"] = "Left",
            ["right"] = "Right",
        },
        order = 150,
    },
    rows = {
        type = "range",
        name = lang["Rows"],
        desc = lang["Sets result row count.\n|cffff0000REQUIRES RELOADUI|r"],
        min = 5,
        max = 25,
        step = 1,
        isPercent = false,
        get = function() return SchoolUI.db.profile.resultRows end,
        set = function(v) SchoolUI.db.profile.resultRows = v end,
        order = 160,
    },
    backgroundalpha = {
        type = "range",
        name = lang["Backdrop Alpha"],
        desc = lang["Sets frame background transparency level."],
        min = 0,
        max = 1,
        step = 0,
        isPercent = false,
        get = function() return SchoolUI.db.profile.backgroundAlpha end,
        set = function(v)
            SchoolUI.db.profile.backgroundAlpha = v
            SchoolUI:SetBackdropAlpha()
        end,
        order = 170,
    },
    
    showin = {
        type = 'group',
        name = lang["Show frame"],
        desc = lang["Show School frame"],
        args = {
            spacerTop = {
                type = "header",
                name = lang["Show frame"],
                order = 1,
            },
            
            login = {
                type = "toggle",
                name = lang["Login"],
                desc = lang["Login"],
                get = function() return SchoolUI.db.profile.showLogin end,
                set = function(v) 
                    SchoolUI.db.profile.showLogin = v
                    SchoolUI:UpdateResults()
                end,
                order = 100,
            },
            
            combat = {
                type = "toggle",
                name = lang["Combat"],
                desc = lang["Combat"],
                get = function() return SchoolUI.db.profile.showCombat end,
                set = function(v)
                    SchoolUI.db.profile.showCombat = v
                    SchoolUI:UpdateResults()
                end,
                order = 110,
            },
            
            talents = {
                type = "toggle",
                name = lang["Talents"],
                desc = lang["Talents"],
                get = function() return SchoolUI.db.profile.showTalents end,
                set = function(v) 
                    SchoolUI.db.profile.showTalents = v 
                    SchoolUI:UpdateResults()
                end,
                order = 120,
            },
            
        },
    },
    
}

SchoolUI:SetDebugging(School:IsDebugging())
SchoolUI:SetDebugLevel(School:GetDebugLevel())

SchoolUI.revision = tonumber(REVISION)

function SchoolUI:OnInitialize()
	self.resultRows = self.db.profile.resultRows
    
    local args = AceLibrary("FuBarPlugin-2.0"):GetAceOptionsDataTable(SchoolUI)

    for k, v in pairs(args) do
        if SchoolOptions.OnMenuRequest.args[k] == nil then
            SchoolOptions.OnMenuRequest.args[k] = v
        end
    end
end

function SchoolUI:OnEnable()
    self:RegisterEvent("CHARACTER_POINTS_CHANGED")
    self:RegisterEvent("PLAYER_LEVEL_UP")
    self:RegisterEvent("PLAYER_MONEY")
    -- self:RegisterEvent("SPELLS_CHANGED")
    
    -- self:RegisterEvent("School_Teacher_Changed")
    -- self:RegisterEvent("School_Student_Changed")
    self:RegisterEvent("School_Changed")

    self:RegisterEvent("School_ShowUI", "ShowUI")
    self:RegisterEvent("School_HideUI", "HideUI")
    self:RegisterEvent("School_OpenUI", "OpenUI")
    
   	self:RegisterEvent("PLAYER_REGEN_ENABLED")

    if (self.db.profile.showLogin) and (self.db.char.frameOpen) then self:ShowUI() end
end

function SchoolUI:OnDisable()
    self.schoolResults = nil
end

function SchoolUI:CHARACTER_POINTS_CHANGED()
    -- self:Debug("CHARACTER_POINTS_CHANGED")
    if (not UnitAffectingCombat("player")) then
        if (not self:IsShown()) then self:ShowUI() end
    end

    if (self:IsShown()) then self:UpdateResults() end
end

function SchoolUI:PLAYER_REGEN_ENABLED()
    if (self.updatePending) then
        self.updatePending = false
        if (self:IsShown()) then self:UpdateResults() end
    end
end

function SchoolUI:PLAYER_LEVEL_UP()
    -- self:Debug("PLAYER_LEVEL_UP")
    if (not UnitAffectingCombat("player")) then
        if (not self:IsShown()) then self:ShowUI() end
    else
        self.updatePending = true
    end

    if (self:IsShown()) then self:UpdateResults() end
end

function SchoolUI:PLAYER_MONEY()
    -- self:Debug("PLAYER_MONEY")
    if (self:IsShown()) then self:UpdateResults() end
end

function SchoolUI:SPELLS_CHANGED()
    -- self:Debug("SPELLS_CHANGED")
    if (self:IsShown()) then
        self:UpdateResults()
    end
end

function SchoolUI:School_Changed()
    -- self:Debug("School_Changed")
    if (self:IsShown()) then
        self:UpdateResults()
    end
end

function SchoolUI:OpenUI()
    if (not self:IsShown()) then
        self:ShowUI()
    else
        self:HideUI()
    end
end

function SchoolUI:ShowUI()
    self:CreateUI()

    self.db.char.frameOpen = true
    if (not self:IsShown()) then
        self.frames.main:Show()
    end
end

function SchoolUI:HideUI()
    self.db.char.frameOpen = false
    if (self:IsShown()) then
        self.frames.main:Hide()
    end
end

function SchoolUI:IsShown()
    if (not self.frames) then return false end
    return self.frames.main:IsShown()
end

function SchoolUI:PlaceTooltip()
	if self.db.profile.tooltipPosition == "left" then
		self.tooltips.show.icon:ClearAllPoints()
		self.tooltips.show.icon:SetPoint("TOPLEFT", self.tooltips.show, "TOPLEFT", -37, -3)

		if self.db.profile.fixedTooltip then
			self.tooltips.show:SetOwner(self.frames.main.items[1], "ANCHOR_BOTTOMLEFT", -2, 43)
		else
			self.tooltips.show:SetOwner(this, "ANCHOR_BOTTOMLEFT")
		end
	else
		self.tooltips.show.icon:ClearAllPoints()
		self.tooltips.show.icon:SetPoint("TOPRIGHT", self.tooltips.show, "TOPRIGHT", 37, -3)

		if self.db.profile.fixedTooltip then
			self.tooltips.show:SetOwner(self.frames.main.items[1], "ANCHOR_BOTTOMRIGHT", 20, 43)
		else
			self.tooltips.show:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
		end
	end
end

function SchoolUI:CreateUI()
    -- self:Debug("CreateUI")
	if self.frames then 
        self:UpdateResults()
        return
    end
    
	self.frames = {}
	self.tooltips = {}
    
    local frameWidth = 300
    local buttonWidth = frameWidth - 10
    
  	local f = CreateFrame("Frame", nil, UIParent)
	f:Hide()
	f:SetFrameStrata("BACKGROUND")
    f:SetWidth(frameWidth)
	f:SetHeight(29 + 19 * self.resultRows)
	f:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = { left = 5, right = 5, top = 5, bottom = 5 },
	})
	f:SetClampedToScreen(true)
	f:SetMovable(true)
	f:EnableMouse(true)
	f:RegisterForDrag("LeftButton") 
	f:SetScript("OnDragStart", function()
		if not self.db.profile.locked then
			this:StartMoving()
		end
	end)
	f:SetScript("OnDragStop", function()
		if not self.db.profile.locked then
			this:StopMovingOrSizing()
			self:SavePosition()
		end
	end)
	f:SetPoint("CENTER", UIParent, "CENTER", 0, 0)

    f.menuButton = CreateFrame("Button", nil, f)
	f.menuButton:SetWidth(buttonWidth)
	f.menuButton:SetHeight(19)
	f.menuButton:RegisterForDrag("LeftButton")
	f.menuButton:SetScript("OnDragStart", function()
		if not self.db.profile.locked then
			this:GetParent():StartMoving()
		end
	end)
	f.menuButton:SetScript("OnDragStop", function()
		if not self.db.profile.locked then
			this:GetParent():StopMovingOrSizing()
			self:SavePosition()
		end
	end)
    f.menuButton:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	f.menuButton:SetScript("OnClick", function()
        if (dewdrop:IsOpen(f)) then dewdrop:Close() end
        if arg1 == "LeftButton" then
            if (not self.schoolResults) then return end
			dewdrop:Open(f, 'children', function()
                dewdrop:AddLine(
                    'text', lang["Hide school frame"],
                    'tooltipTitle', lang["Hide"],
                    'tooltipText', lang["Hide school frame"],
                    'func', function()
                        self:TriggerEvent("School_HideUI")
                        dewdrop:Close()
                    end
                )
                    
                dewdrop:AddLine(
                    'isTitle', true,
                    'isClickable', false
                )
                
                dewdrop:AddLine(
                    'isTitle', true,
                    'text', "School Info"
  				)

                dewdrop:AddLine(
                    'isTitle', true,
                    'isClickable', false
                )

                dewdrop:AddLine(
                    'isTitle', true,
                    'text', "Spells"
  				)

                if (self.schoolResults.totals.spells.known > 0) then
                    dewdrop:AddLine(
                        'text', "Known: " .. self.schoolResults.totals.spells.known,
                        'isClickable', false
                    )
                end

                dewdrop:AddLine(
                    'text', "Learnable: " .. self.schoolResults.totals.spells.met,
                    'isClickable', false
                )

                dewdrop:AddLine(
                    'text', "Total: " .. self.schoolResults.totals.spells.all,
                    'isClickable', false
                )
                
                dewdrop:AddLine(
                    'isTitle', true,
                    'isClickable', false
                )

                dewdrop:AddLine(
                    'isTitle', true,
                    'text', "Cost"
  				)
                
                if (self.schoolResults.totals.cost.known > 0) then
                    dewdrop:AddLine(
                        'text', "Known: " .. abacus:FormatMoneyExtended(self.schoolResults.totals.cost.known, true, true),
                        'isClickable', false
                    )
                end

                dewdrop:AddLine(
                    'text', "Learnable: " .. abacus:FormatMoneyExtended(self.schoolResults.totals.cost.met, true, true),
                    'isClickable', false
                )
                
                dewdrop:AddLine(
                    'text', "Total: " .. abacus:FormatMoneyExtended(self.schoolResults.totals.cost.all, true, true),
                    'isClickable', false
                )
			end)
        else
			dewdrop:Open(f, 'children', function()
				dewdrop:FeedAceOptionsTable(self.core.options)
			end)
        end
        
	end)
 	f.menuButton:SetPoint("TOPLEFT", f, "TOPLEFT", 5, -5)

	f.headerBar = f:CreateTexture(nil, "ARTWORK")
	f.headerBar:SetWidth(buttonWidth)
	f.headerBar:SetHeight(19)
	f.headerBar:SetTexture("Interface\\AddOns\\School\\Textures\\otravi.tga")
	f.headerBar:SetPoint("TOPLEFT", f, "TOPLEFT", 5, -5)
	f.headerBar:SetGradient("HORIZONTAL", 0.8, 0, 0, 0.4, 0, 0)

	f.headerText = f:CreateFontString(nil, "OVERLAY")
	f.headerText:SetWidth(buttonWidth)
	f.headerText:SetHeight(12)
	f.headerText:SetPoint("TOPLEFT", f, "TOPLEFT", 6, -8)
	f.headerText:SetFontObject(GameFontNormal)
	f.headerText:SetJustifyH("LEFT")
	f.headerText:SetTextColor(1, 1, 1)
	f.headerText:SetText("School")

    --[[
	f.closeButton = CreateFrame("Button", nil, f, "UIPanelCloseButton")
	f.closeButton:SetWidth(27)
	f.closeButton:SetHeight(27)
	f.closeButton:SetPoint("TOPRIGHT", f) 
    --]]
    
	f.scroll = CreateFrame("ScrollFrame", "School_ItemsScrollFrame", f, "FauxScrollFrameTemplate")
	f.scroll:SetScript("OnVerticalScroll", function()
		FauxScrollFrame_OnVerticalScroll(19, function()
			self:UpdateResults()
		end)
	end)
	f.scroll:SetPoint("TOPLEFT", f, "TOPLEFT", 0, -23)
	f.scroll:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -28, 5)

	self.frames.main = f

	self.frames.main.items = {}

	local p
    
    local detailTooltip = {}
    if (self.db.profile.detailPosition == "left") then
        detailTooltip.point = "BOTTOMRIGHT"
        detailTooltip.relativePoint = "BOTTOMLEFT"
    else
        detailTooltip.point = "BOTTOMLEFT"
        detailTooltip.relativePoint = "BOTTOMRIGHT"
    end
    
	for i = 1, self.resultRows do
		local f = CreateFrame("Button", nil, self.frames.main)
		f.schoolID = nil
		f:SetWidth(buttonWidth)
		f:SetHeight(18)
		f:RegisterForClicks("LeftButtonUp", "RightButtonUp")
		f:SetScript("OnEnter", function()
			if this:IsVisible() then
				self:ShowItemTooltip(this.schoolID, this.icon:GetTexture())
				this.highlight:Show()
			end
		end)
		f:SetScript("OnLeave", function()
			self.tooltips.show:Hide()
			this.highlight:Hide()
		end)

		f:SetScript("OnClick", function()
			if this:IsVisible() then
				if arg1 == "LeftButton" then
					if dewdrop:IsOpen(f) then
						dewdrop:Close()
					else
                        if (not self.schoolResults) then return end
                        local item = self.schoolResults.data[f.schoolID]

                        dewdrop:Open(self.frames.main, 'point', detailTooltip.point, 'relativePoint', detailTooltip.relativePoint, 'children', function()
                            dewdrop:AddLine(
								'isTitle', true,
                                'text', item.name,
								'icon', f.icon:GetTexture()
							)
                            
                            if (item.subtext ~= nil) and (item.subtext ~= "None") then
                                dewdrop:AddLine(
                                    'text', crayon:Green(item.subtext),
                                    'isClickable', false
                                )
                            end
                            
                            --[[
                            if (item.description ~= nil) and (item.desciption ~= "") then
                                dewdrop:AddLine(
                                    'text', item.description,
                                    'isClickable', false
                                )
                            end
                            --]]
                            
                            dewdrop:AddLine(
                                'text', lang["Cost"] .. ": " .. abacus:FormatMoneyExtended(item.cost.money, true),
                                'isClickable', false
                            )
                            
                            if (GetMoney() < item.cost.money) and (not item.known) then
                                dewdrop:AddLine(
                                    'text', crayon:Red(lang["You can't afford this."]),
                                    'isClickable', false
                                )
                            end
                            
                            -- if (UnitLevel("player") < item.level) and (not item.known) then
                            -- if (not item.met) and (not item.known) then
                            if (not item.criteria.level.met) and (not item.known) then
                                local lowLevel = lang["Your level is too low."]
                                local requiresLevel = string.format(lang["Requires level %d"], item.level)
                                
                                local msgLevel = ""
                                -- msgLevel = lowLevel .. " "
                                msgLevel = msgLevel .. requiresLevel
                                
                                dewdrop:AddLine(
                                    'text', crayon:Purple(msgLevel),
                                    'isClickable', false
                                )
                            end
                            
                            if (not item.criteria.skill.met) then
                                local lowSkill = "%s"
                                local lowRank = " (%d)"
                                
                                local noSkill = ""
                                if (item.criteria.skill.name) then noSkill = string.format(lowSkill, item.criteria.skill.name) end
                                if (item.criteria.skill.rank) and (item.criteria.skill.rank > 0) then 
                                    noSkill = noSkill .. string.format(lowRank, item.criteria.skill.rank)
                                end
                                noSkill = string.format(lang["Requires %s"], noSkill)
                                noSkill = crayon:Purple(noSkill)
                                
                                dewdrop:AddLine(
                                    'text', noSkill,
                                    'isClickable', false
                                )
                            end
                            
                            if (item.known) then
                                dewdrop:AddLine(
                                    'text', crayon:Silver(lang["Already known"]),
                                    'isClickable', false
                                )
                            end
                            
                            dewdrop:AddLine(
                                'isTitle', true,
                                'isClickable', false
                            )
                            
                            if (item.trainers ~= nil) then
                                dewdrop:AddLine(
                                    'isTitle', true,
                                    'text', lang["Trainers"],
                                    'isClickable', false
                                )
                                for tk, tv in pairs(item.trainers) do
                                    if (tv ~= nil) then
                                        local trainer = ""
                                        if (not tv.name) then trainer = lang["Unknown"]
                                        else trainer = tv.name end
                                        
                                        --[[
                                        if (not tv.subzone) or (tv.subzone == "") then trainer = trainer .. " " .. lang["at"] .. " " .. lang["Unknown"]
                                        else trainer = trainer .. " " .. lang["at"] .. " " .. tv.subzone end
                                        if (not tv.zone) then trainer = trainer .. " " .. lang["in"] .. " " .. lang["Unknown"]
                                        else trainer = trainer .. " " .. lang["in"] .. " " .. tv.zone end
                                        --]]
                                        
                                        if (tv.subzone) and (tv.subzone ~= "") then
                                            trainer = trainer .. " " .. lang["at"] .. " " .. tv.subzone
                                        end
                                        
                                        if (tv.zone) and (tv.zone ~= "") then 
                                            trainer = trainer .. " " .. lang["in"] .. " " .. tv.zone
                                        end

                                        if (tv.coords) then
                                            trainer = trainer .. string.format(" (%.0f,%.0f)", tv.coords.x * 100, tv.coords.y * 100)
                                        end
                                        
                                        dewdrop:AddLine(
                                            'text', trainer,
                                            'isClickable', false
                                        )
                                    end
                                end

                                dewdrop:AddLine(
                                    'isTitle', true,
                                    'isClickable', false
                                )
                            end
                            
							dewdrop:AddLine(
								'text', lang["Close window"],
								'func', function() dewdrop:Close() end
							)
						end)
					end
				end
			end
		end)

		f.icon = f:CreateTexture(nil, "OVERLAY")
		f.icon:SetWidth(18)
		f.icon:SetHeight(18)
		f.icon:SetPoint("LEFT", f, "LEFT", 0, 0)

		f.highlight = f:CreateTexture(nil, "ARTWORK")
		f.highlight:SetWidth(buttonWidth)
		f.highlight:SetHeight(18)
		f.highlight:SetTexture("Interface\\HelpFrame\\HelpFrameButton-Highlight")
		f.highlight:SetPoint("TOPLEFT", f, "TOPLEFT", 0, 0)
		f.highlight:SetTexCoord(0, 1, 0, 0.578125)
		f.highlight:Hide()

		f.name = f:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		f.name:SetWidth(buttonWidth)
		f.name:SetHeight(18)
		f.name:SetJustifyH("LEFT")
		f.name:SetPoint("LEFT", f, "LEFT", 19, 0)

		if i == 1 then
			f:SetPoint("TOPLEFT", self.frames.main, "TOPLEFT", 6, -25)
		else
			f:SetPoint("TOPLEFT", p, "BOTTOMLEFT", 0, -1)
		end

		table.insert(self.frames.main.items, f)
		p = f
	end
    local f = CreateFrame("GameTooltip", "School_TooltipShow", UIParent, "GameTooltipTemplate")

	f.icon = f:CreateTexture(nil, "ARTWORK")
	f.icon:SetWidth(39)
	f.icon:SetHeight(39)

	self.tooltips.show = f

	self:SetBackdropAlpha()
	self:RestorePosition()
    self:UpdateResults()
end

function SchoolUI:SetBackdropAlpha()
	if not self.frames then return end

	self.frames.main:SetBackdropColor(0, 0, 0, self.db.profile.backgroundAlpha)
end

function SchoolUI:SavePosition()
	if not self.frames then return end

	local x, y, s = self.frames.main:GetLeft(), self.frames.main:GetTop(), self.frames.main:GetEffectiveScale()

	x, y = x * s, y * s

	self.db.profile.mainPositionX = x
	self.db.profile.mainPositionY = y
end

function SchoolUI:RestorePosition()
	if not self.frames then return end

	local x, y, s = self.db.profile.mainPositionX, self.db.profile.mainPositionY, self.frames.main:GetEffectiveScale()

	if not x or not y then
		return
	end

	x, y = x / s, y / s

	self.frames.main:ClearAllPoints()
	self.frames.main:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x, y)
end

function SchoolUI:ShowItemTooltip(schoolID, icon)
    if (not self.schoolResults) or (not self.schoolResults.data) then return end
    if (not self.schoolResults.data[schoolID]) then return end
    
    local result = self.schoolResults.data[schoolID]
    local resultInfo = result.name
    if (result.subtext ~= nil) and (result.subtext ~= "None") then
        resultInfo = resultInfo .. "\n" .. crayon:Green(result.subtext)
    end

    --[[
    if (result.description ~= nil) and (result.description ~= "") then
        resultInfo = resultInfo .. "\n\n" .. result.description .. "\n"
    end
    --]]
    
    resultInfo = resultInfo .. "\n" .. lang["Category"] .. ": " .. result.category .. "\n" .. lang["Cost"] .. ": " .. abacus:FormatMoneyShort(result.cost.money, true)
    if (not result.known) then
        if (GetMoney() < result.cost.money) then
            resultInfo = resultInfo .. "\n" .. crayon:Red(lang["You can't afford this."])
        end
        
        --[[
        if (UnitLevel("player") < result.level) then
            resultInfo = resultInfo .. "\n" .. crayon:Purple(lang["Your level is too low."])
            resultInfo = resultInfo .. "\n" .. crayon:Purple(string.format(lang["Requires level %d"], result.level))
        end
        --]]
        if (not result.criteria.level.met) then
            resultInfo = resultInfo .. "\n" .. crayon:Purple(lang["Your level is too low."])
            resultInfo = resultInfo .. "\n" .. crayon:Purple(string.format(lang["Requires level %d"], result.level))
        end
        
        if (not result.criteria.skill.met) then
            local noSkill = ""
            if (result.criteria.skill.name) then noSkill = noSkill .. result.criteria.skill.name end
            if (result.criteria.skill.rank) and (result.criteria.skill.rank > 0) then  noSkill = noSkill .. " (" .. result.criteria.skill.rank .. ")" end
            noSkill = string.format(lang["Requires %s"], noSkill)
            
            resultInfo = resultInfo .. "\n" .. crayon:Purple(noSkill)
        end
    else
        resultInfo = resultInfo .. "\n" .. crayon:Silver(lang["Already known"])
    end
    self:PlaceTooltip()
    self.tooltips.show:SetOwner(UIParent, "ANCHOR_PRESERVE")
    self.tooltips.show:SetText(resultInfo)
    self.tooltips.show:Show()
    self.tooltips.show.icon:SetTexture(icon)
end

function SchoolUI:FetchResults()
    -- self:Debug("FetchResults")
    -- return School:Clone(School:FetchResults())
    return School:FetchResults()
end

function SchoolUI:UpdateResults()
    -- self:Debug("UpdateResults")
	if not self.frames then return end

    self.schoolResults = self:FetchResults()
    if (not self.schoolResults) then self.schoolResults = {} end
    if (not self.schoolResults.data) then self.schoolResults.data = {} end
    if (not self.schoolResults.count) then self.schoolResults.count = 0 end
    
    self.studentLevel = tonumber(UnitLevel("player"))
    self.studentMoney = GetMoney()
    self.numResults = self.schoolResults.count
    self.totalCost = self.schoolResults.totalcost

	FauxScrollFrame_Update(self.frames.main.scroll, self.numResults, self.resultRows, 19);

    local colors = {}
    colors.icons = {}
    colors.icons.normal = {1,1,1}
    colors.icons.known = {0,1,0}
    colors.icons.nolevel = {153,0,128}
    colors.icons.nomoney = {1,0,0}
    
    local chrTab = "    "
    
	for i = 1, self.resultRows do
		local index = i + FauxScrollFrame_GetOffset(self.frames.main.scroll)

        -- self:Debug("index ==> %s\nself.numResults ==> %s", index, self.numResults)
        if (index <= self.numResults) then
            local item = self.schoolResults.data[index]
            self.frames.main.items[i].schoolID = index
            self.frames.main.items[i].icon:SetTexture(item.icon)
            local fullname = item.fullname
            local money = item.cost.money            
            if (self.studentMoney < item.cost.money) then money = crayon:Red(abacus:FormatMoneyFull(money, false, false))
            else money = abacus:FormatMoneyFull(money, true) end
            -- if (self.studentLevel < item.level) then fullname = crayon:Purple(item.fullname) end
            if (not item.met) then fullname = crayon:Purple(item.fullname) end
            if (item.known) then fullname = crayon:Silver(item.fullname) end
            
            local line = fullname
            if (not item.known) then line = line .. chrTab .. money end
            self.frames.main.items[i].name:SetText(line)

            -- self.frames.main.items[i].icon:SetVertexColor(1,1,1)
            self.frames.main.items[i].icon:SetVertexColor(
                colors.icons.normal[1],
                colors.icons.normal[2],
                colors.icons.normal[3]
            )
            
            --[[
            if (self.studentMoney < item.cost.money) and (not item.known) then
                self.frames.main.items[i].icon:SetVertexColor(
                    colors.icons.nomoney[1],
                    colors.icons.nomoney[2],
                    colors.icons.nomoney[3]
                )
            end
            --]]
            
            --[[
            if (self.studentLevel < item.level) then
                self.frames.main.items[i].icon:SetVertexColor(
                    colors.icons.nolevel[1],
                    colors.icons.nolevel[2],
                    colors.icons.nolevel[3]
                )
            end
            --]]
            
            --[[
            if (item.known) then
                self.frames.main.items[i].icon:SetVertexColor(
                    colors.icons.known[1],
                    colors.icons.known[2],
                    colors.icons.known[3]
                )
            end
            --]]
            
			self.frames.main.items[i]:Show()
		else
			self.frames.main.items[i]:Hide()
		end
	end

    local headerText = ""
    local headerName = SchoolLocale.school["School"]
    local headerNumResults = ""
    local headerTotalCost = ""
    if (self.numResults) then headerNumResults = string.format(" (%d)", self.numResults) end
    if (self.totalCost) then
        local totalCost = 0
        if (School.db.profile.totalCostAll) then totalCost = self.totalcost.all end
        if (School.db.profile.totalCostKnown) then totalCost = self.totalcost.known end
        if (School.db.profile.totalCostMet) then totalCost = self.totalcost.met end
        
        if (School.db.profile.totalCost) then
            local totalCost = self.totalCost
            if (self.studentMoney < self.totalCost) then totalCost = crayon:Red(abacus:FormatMoneyExtended(totalCost))
            else totalCost = abacus:FormatMoneyExtended(totalCost, true) end
            headerTotalCost = string.format(" -- %s", totalCost)
        end
    end

    headerText = headerName
    if (headerNumResults) then headerText = headerText .. headerNumResults end
    if (headerTotalCost) then headerText = headerText .. headerTotalCost end
    
	-- self.frames.main.headerText:SetText(string.format("School (%d)", self.numResults))
    self.frames.main.headerText:SetText(headerText)
end
    