assert(School, "School not found!")

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local lang = AceLibrary("AceLocale-2.2"):new("SchoolOptions") 
local tablet = AceLibrary("Tablet-2.0") 

-------------------------------------------------------------------------------
-- Localization
-------------------------------------------------------------------------------

lang:RegisterTranslations("enUS", function() return {
	["|cffeda55fClick|r to open/close School frame."] = true,
	["minimap"] = true,
	["Minimap"] = true,
	["Toggle the minimap button."] = true,
	["Hidden"] = true,
	["Shown"] = true,
} end)

-------------------------------------------------------------------------------
-- Main
-------------------------------------------------------------------------------

local deuce = School:NewModule("Options")
deuce.internalModule = true
deuce.consoleCmd = not FuBar and lang["minimap"]
deuce.consoleOptions = not FuBar and {
	type = "toggle",
	name = lang["Minimap"],
	desc = lang["Toggle the minimap button."],
	get = function() return SchoolOptions.minimapFrame and SchoolOptions.minimapFrame:IsVisible() or false end,
	set = function(v) if v then SchoolOptions:Show() else SchoolOptions:Hide() end end,
	map = {[false] = lang["Hidden"], [true] = lang["Shown"]},
	hidden = function() return FuBar and true end,
} 

-------------------------------------------------------------------------------
-- FuBar Plugin
-------------------------------------------------------------------------------

SchoolOptions = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
SchoolOptions.name = "FuBar - School"
SchoolOptions:RegisterDB("SchoolFubarDB")

-- SchoolOptions.hasIcon = "Interface\\School\\Icons\\School_Default"
SchoolOptions.hasIcon = "Interface\\Icons\\INV_Misc_Book_08"
SchoolOptions.defaultMinimapPosition = 250
SchoolOptions.clickableTooltip = true 
SchoolOptions.cannotDetachTooltip = true
SchoolOptions.defaultPosition = "LEFT"
SchoolOptions.tooltipHiddenWhenEmpty = false
SchoolOptions.hideWithoutStandby = true
SchoolOptions.OnMenuRequest = deuce.core.options

local args = AceLibrary("FuBarPlugin-2.0"):GetAceOptionsDataTable(SchoolOptions)
for k, v in pairs(args) do
	if SchoolOptions.OnMenuRequest.args[k] == nil then
		SchoolOptions.OnMenuRequest.args[k] = v
	end
end

-------------------------------------------------------------------------------
-- FuBar Methods
-------------------------------------------------------------------------------

function SchoolOptions:OnEnable()
    if (SchoolOptions.fullFubar) then
        -- SchoolOptions:RegisterEvent("School_Changed", "UpdateResults", 5)
        SchoolOptions:RegisterEvent("School_Changed", "Update", 5)
        SchoolOptions:Update()
    end
end

function SchoolOptions:OnDisable()
    if (SchoolOptions:IsEventRegistered("School_Changed")) then
        SchoolOptions:UnregisterEvent("School_Changed")
    end
end

function SchoolOptions:OnClick()
	self:TriggerEvent("School_OpenUI")
end

function SchoolOptions:OnTooltipUpdate()
    if (SchoolOptions.fullFubar) then self:UpdateResults() end
    if (SchoolUI) then tablet:SetHint(lang["|cffeda55fClick|r to open/close School frame."]) end
end
