local _, _, REVISION = string.find("$Revision: 23349 $", "(%d+)")

if not AceLibrary then error("School requires AceLibrary") end
if not AceLibrary:HasInstance("Toolbox-1.0") then error("School requires Toolbox-1.0") end

-------------------------------------------------------------------------------
-- Localization
-------------------------------------------------------------------------------
SchoolLocale = {}
SchoolLocale.school = AceLibrary("AceLocale-2.2"):new("School")

SchoolLocale.school:RegisterTranslations("enUS", function() return {
    ["SLASHCMD_LONG"] = "/school",
    ["SLASHCMD_SHORT"] = "/school",

    ["Change debug level"] = true,
    ["Cost"] = true,

    ["Debug level"] = true,
    ["Debug settings"] = true,

    ["Enabled"] = true,
    ["Enable/disable debugging"] = true,
    ["Enable or disable this module."] = true,

    ["Ignore cost"] = true,
    ["Ignore combat"] = true,
    ["Ignore level modifier"] = true,
    ["Ignore level override"] = true,
    ["Ignore player level modifier"] = true,
    ["Ignore player level override"] = true,
    ["Ignore skill level"] = true,
    ["Ignore skill level override"] = true,
    ["Include known"] = true,
    ["Include known spells/skills."] = true,
    

    ["Level Modifiers"] = true,
    ["Level modifier settings."] = true,
    ["Level modifier value"] = true,
    
    ["Level Overrides"] = true,
    ["Level override settings."] = true,
    ["Level override value"] = true,

    ["Maximum player level."] = true,
    ["Maximum skill level."] = true,
    
    ["Number to add to your level."] = true,
    
    ["Options for %s module."] = true,

    ["Player"] = true,
    ["Player level modifier value"] = true,
    ["Player level override value"] = true,

    ["Rebuild"] = true,
	["Rebuild %s data."] = true,    
    ["Rebuild categories."] = true,
    ["Rebuild classes."] = true,
    ["Rebuild data."] = true,
    ["Rebuild profile."] = true,
    ["Rebuild races."] = true,
    ["Rebuild school."] = true,
    ["Rebuild student."] = true,
    ["Rebuild teacher."] = true,
    ["Rebuild trainers."] = true,

    ["School"] = true,
    ["School Commands"] = true,
    ["School Help"] = true,
    ["School Options"] = true,
    ["Show All"] = true,
    ["Show cost totals"] = true,
    ["Skill level override value"] = true,
    ["Skill level value"] = true,
    ["Skill"] = true,

    ["Toggle debugging"] = true,
    ["Toggle ignore combat."] = true,
    ["Toggle ignore learning cost."] = true,
    ["Toggle ignore level modifier."] = true,
    ["Toggle ignore level override."] = true,
    ["Toggle ignore player level modifier."] = true,
    ["Toggle ignore player level override."] = true,
    ["Toggle ignore skill level."] = true,
    ["Toggle ignore skill level override."] = true,
    ["Toggle show cost totals."] = true,
 
    ["What do you want to learn today?"] = true,
 
} end)

-------------------------------------------------------------------------------
-- Declaration
-------------------------------------------------------------------------------

if AceLibrary:HasInstance("AceDebug-2.0") then
    School = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "AceModuleCore-2.0", "AceDebug-2.0", "Toolbox-1.0");
    School:SetDebugging(false)
    School:SetDebugLevel(3)
else
	School = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "AceModuleCore-2.0", "Toolbox-1.0");
	function School:Debug() end
    function School:LevelDebug() end
    function School:IsDebugging() return false end
    function School:SetDebugging() end
    function School:GetDebugLevel() end
    function School:SetDebugLevel() end
end

School.revision = tonumber(REVISION)
School.maxRevision = School.revision
School.dbversion = 2

School.debugFrame = ChatFrame4

local lang = SchoolLocale.school

School.defaults = {
    ignoreOverrideLevelPlayer = true,
    ignoreOverrideLevelSkill = true,
    ignoreCost = true,
    totalCost = true,
    
    -- Debug
    overrideLevelPlayer = 20,
    ignoreOverrideLevelPlayer = false,
    
}

School.options = {
    type = 'group',
    name = lang["School"],
    desc = lang["What do you want to learn today?"],
    args = {
		spacerTop = {
			type = "header",
			order = 1,
		},
        
        includeknown = {
            type = "toggle",
            name = lang["Include known"],
            desc = lang["Include known spells/skills."],
            get = function() return School.db.profile.includeKnown end,
            set = function(v)
                School.db.profile.includeKnown = v
                School:GotoSchool()
            end,
            order = 50,
        },
        
        leveloverrides = {
            type = 'group',
            name = lang["Level Overrides"],
            desc = lang["Level override settings."],
            args = {
        		spacerTop = {
        			type = "header",
                    name = lang["Level Overrides"],
        			order = 1,
        		},                

           		spacerPlayer = {
        			type = "header",
                    name = lang["Player"],
        			order = 100,
                    -- hidden = true,
        		},                

                player = {
                    type = "text",
                    name = lang["Player level override value"],
                    desc = lang["Maximum player level."],
                    usage = "<level>",
                    get = function() return School.db.profile.overrideLevelPlayer end,
                    set = function(v)
                        School.db.profile.overrideLevelPlayer = tonumber(v)
                        -- School:Debug("v: %s\nlevel: %s", v, UnitLevel("player"))
                        -- if (tonumber(v) < UnitLevel("player")) then School.db.profile.ignoreOverrideLevelPlayer = true 
                        if (School.db.profile.overrideLevelPlayer < UnitLevel("player")) then School.db.profile.ignoreOverrideLevelPlayer = true
                        else School.db.profile.ignoreOverrideLevelPlayer = false end
                        School:GotoSchool()
                    end,
                    validate = function(v)
                        if (type(v) ~= "number") then v = tonumber(v) end
                        if (not v) then return false end
                        if (v > 70) then return false end
                        return true
                    end,
                    order = 101,
                },
                
                
                ignoreplayer = {
                    type = "toggle",
                    name = lang["Ignore player level override"],
                    desc = lang["Toggle ignore player level override."],
                    get = function() return School.db.profile.ignoreOverrideLevelPlayer end,
                    set = function(v) 
                        School.db.profile.ignoreOverrideLevelPlayer = v
                        School:GotoSchool()
                    end,
                    order = 102
                },

        		spacerSkill = {
        			type = "header",
                    name = lang["Skill"],
        			order = 120,
        		},                
                
                skill = {
                    type = "text",
                    name = lang["Skill level value"],
                    desc = lang["Maximum skill level."],
                    usage = "<level>",
                    get = function() return School.db.profile.overrideLevelSkill end,
                    set = function(v)
                        School.db.profile.overrideLevelSkill = tonumber(v)
                        if (School.db.profile.overrideLevelSkill < 1) then School.db.profile.ignoreOverrideLevelSkill = true
                        else School.db.profile.ignoreOverrideLevelSkill = false end
                        School:GotoSchool()
                    end,
                    validate = function(v)
                        if (type(v) ~= "number") then v = tonumber(v) end
                        if (not v) then return false end
                        -- if (v > 450) then return false end
                        return true
                    end,
                    order = 121,
                },
                
                ignoreskill = {
                    type = "toggle",
                    name = lang["Ignore skill level"],
                    desc = lang["Toggle ignore skill level."],
                    get = function() return School.db.profile.ignoreOverrideLevelSkill end,
                    set = function(v) 
                        School.db.profile.ignoreOverrideLevelSkill = v
                        School:GotoSchool()
                    end,
                    order = 122
                },
            },
        },
    },
    handler = School
}

School.options.args.standby = nil

School:RegisterChatCommand({lang["SLASHCMD_LONG"], lang["SLASHCMD_SHORT"]}, School.options)

School:RegisterDB("SchoolDB", "SchoolCharDB")
School:RegisterDefaults("profile", School.defaults)

if AceLibrary:HasInstance("AceDB-2.0") then
    School.charid = AceLibrary("AceDB-2.0").CHAR_ID
    School.realmid = AceLibrary("AceDB-2.0").REALM_ID
    School.classid = AceLibrary("AceDB-2.0").CLASS_ID
    School.faction = AceLibrary("AceDB-2.0").FACTION
    School.realm = AceLibrary("AceDB-2.0").REALM
    School.name = AceLibrary("AceDB-2.0").NAME
else
    School.charid = ""
    School.realmid = ""
    School.classid = ""
    School.faction = ""
    School.realm = ""
    School.name = UnitName("player")
end

School.data = {}
School.modules = {}

-------------------------------------------------------------------------------
-- Module prototype 
-------------------------------------------------------------------------------   

if AceLibrary:HasInstance("AceDebug-2.0") then
    School:SetModuleMixins("Toolbox-1.0", "AceEvent-2.0", "AceDebug-2.0")
else 
    School:SetModuleMixins("Toolbox-1.0", "AceEvent-2.0")
    
    function School.modulePrototype:Debug() end
    function School.modulePrototype:LevelDebug() end
    function School.modulePrototype:IsDebugging() return false end
    function School.modulePrototype:SetDebugging() end
    function School.modulePrototype:GetDebugLevel() end
    function School.modulePrototype:SetDebugLevel() end
end

School.modulePrototype.core = School
School.modulePrototype.revision = 1
School.modulePrototype.debugFrame = School.debugFrame

function School.modulePrototype:Enabled()
	return self.core:IsModuleActive(self)
end

function School.modulePrototype:GetVersion()
	return self.core.maxRevision
end

function School.modulePrototype:InCombat()
	return self.core.inCombat
end

-------------------------------------------------------------------------------
-- Main
-------------------------------------------------------------------------------

function School:OnInitialize()
    -- Do database version check here.
    
	local rev = self.revision

	for name, module in self:IterateModules() do
		if module.revision > rev then
			rev = module.revision
		end

		self:RegisterModule(name, module)
	end

	self.version = self.version .. " |cffff8888r".. rev .."|r" 
	self.maxRevision = rev
    
    if (SchoolOptions.fullFubar) or SchoolUI then
        self.options.args.cost = {
            type = 'group',
            name = lang["Cost"],
            desc = lang["Cost"],
            args = {
                spacerTop = {
                    type = "header",
                    name = lang["Cost"],
                    order = 1,
                },
                
                ignorecost = {
                    type = "toggle",
                    name = lang["Ignore cost"],
                    desc = lang["Toggle ignore learning cost."],
                    get = function() return School.db.profile.ignoreCost end,
                    set = function(v) 
                        School.db.profile.ignoreCost = v
                        School:TriggerEvent("School_Changed")
                    end,
                    order = 100,
                },
                
                totals = {
                    type = "toggle",
                    name = lang["Show cost totals"],
                    desc = lang["Toggle show cost totals."],
                    get = function() return School.db.profile.totalCost end,
                    set = function(v) 
                        School.db.profile.totalCost = v
                        School:TriggerEvent("School_Changed")
                    end,
                    order = 110,
                },
            },
        }
    end

end

function School:OnEnable()
    if (not self.RegisterEvents) and (self.modules.student.eventsLoaded) then self:RegisterEvents() end
end

function School:OnDisable()
    self.data = nil
    self.modules = nil
    
    self:UnregisterEvents()
end

function School:RegisterEvents()
    self:Debug("RegisterEvents")
    if (not self.eventsLoaded) then
        self:RegisterEvent("CHARACTER_POINTS_CHANGED", "GotoSchool")
        self:RegisterEvent("PLAYER_LEVEL_UP", "GotoSchool")
        self:RegisterEvent("SPELLS_CHANGED", "GotoSchool", 2)
    
        self:RegisterEvent("School_TeacherChanged", "GotoSchool")
        self:RegisterEvent("School_StudentChanged", "GotoSchool")
        
        self:RegisterEvent("School_Build", "GotoSchool")
    
        self.inCombat = false

        self:RegisterEvent("PLAYER_REGEN_DISABLED", function() self.inCombat = true end)
        self:RegisterEvent("PLAYER_REGEN_ENABLED", function() self.inCombat = false end)
        self:RegisterEvent("PLAYER_ENTERING_WORLD", function() self.inCombat = false School:GotoSchool() end)    
    end

    self.eventsLoaded = true
    
    self:TriggerEvent("School_Build")
end

function School:UnregisterEvents()
    if (self.eventsLoaded) then
        self:UnregisterAllEvents()
    end

    self.eventsLoaded = false
end

function School:RegisterModule(name, module)
	if module.internalModule then return end
    if module.disabled then return end

	self:RegisterDefaults(name, "profile", module.defaults or {})
	module.db = self:AcquireDBNamespace(name)

	if module.commands then
		for k, v in pairs(module.commands) do
			self.options.args[k] = v
		end
	end

	if not module.notToggable or module.options then
		local m = module
		local options = {
			type = "group",
			name = name,
			desc = string.format(lang["Options for %s module."], name),
			order = 200,
			args = {
				moduleHeader = {
					type = "header",
					name = name,
					order = 1,
				},
			},
		}

		if not module.notToggable then
			options.args.enabled = {
				type = "toggle",
				name = lang["Enabled"],
				desc = lang["Enable or disable this module."],
				get = function() return m.core:IsModuleActive(m) end,
				set = function() m.core:ToggleModuleActive(m) end,
				order = 2,
			}
		end

		if module.options then
			if not module.notToggable then
				options.args.spacerHeader = {
					type = "header",
					order = 3,
				}
			end

			for k, v in pairs(module.options) do
				options.args[k] = v
			end
		end

		self.options.args[string.lower(string.gsub(name, " ", ""))] = options
	end
end    

-------------------------------------------------------------------------------
-- Settings & Commands
-------------------------------------------------------------------------------

function School:GetData()
    -- self:Debug("GetData")
    if (not self.data) then return end
    
    if (self.data.school ~= nil) then
        local count = 0
        for k,v in pairs(self.data.school) do
            -- self:Debug("Key ==> %s", k)
            if (v ~= nil) then
                for vk, vv in pairs(v) do
                    -- self:Debug("Value Key ==> %s", vk)
                    if (vv ~= nil) then
                        for vvk, vvv in pairs(vv) do
                            -- self:Debug("Value Value Key ==> %s", vvk)
                            if (vvv ~= nil) then
                                count = count + 1
                            end
                        end
                    end
                end
            end
        end
        -- self:Debug("self.data.school:count ==> %s", count)
    end
    
    return self.data.school
end

function School:BuildTeacher()
    -- self:Debug("BuildTeacher")
    if (not self.modules.teacher) then return end
    self.data.teacher = self.modules.teacher:GetData()
    
    return self.data.teacher
end

function School:BuildStudent()
    -- self:Debug("BuildStudent")
    if (not self.modules.student) then return end
    self.data.student = self.modules.student:GetData()
    
    return self.data.student
end

function School:BuildSchool()
    -- self:Debug("BuildSchool")
    
    if (not self.data.teacher) then return end
    -- self:Debug("Found Teacher data.")
    
    if (not self.data.student) then return end
    -- self:Debug("Found Student data.")
    
    local school = {}
    local teacher = self.data.teacher
    local student = self.data.student
    
    --[[
    if (teacher.counters) then
        self:Debug("spells: %s\ntrainers: %s\ncharacter spells: %s", teacher.counters.spells, teacher.counters.trainers, teacher.counters.charspells)
    end
    --]]
    
    student.level = UnitLevel("player")
    if (not School.db.profile.ignoreOverrideLevelPlayer) then student.level = School.db.profile.overrideLevelPlayer end
    student.level = tonumber(student.level)
    -- self:Debug("student.level ==> %s", student.level)

    local includeKnown = School.db.profile.includeKnown
    
    if (not teacher.charspells) or (not teacher.spells) then return end
    -- self:Debug("Found Character Spells data.")
    
    local apprentice = {}
    apprentice.primary = {
                          ["Apprentice Alchemist"] = true,
                          ["Apprentice Blacksmith"] = true,
                          ["Apprentice Enchanting"] = true,
                          ["Apprentice Engineer"] = true,
                          ["Apprentice Herbalist"] = true,
                          ["Apprentice Jewelcrafting"] = true,
                          ["Apprentice Miner"] = true,
                          ["Apprentice Leatherworking"] = true,
                          ["Apprentice Skinning"] = true,
                          ["Apprentice Tailor"] = true,
    }
    apprentice.secondary = {
                            ["Apprentice Cook"] = true,
                            ["Apprentice First Aid"] = true,
                            ["Apprentice Fishing"] = true,
    }
    
    for k,v in pairs(teacher.charspells) do
        -- Category
        if (v ~= nil) then
            -- self:Debug("Key ==> %s", k)
            for vk, vv in pairs(v) do
                -- Spell
                if (vv ~= nil) then
                    -- self:Debug("Value Key ==> %s", vk)
                    for vvk, vvv in pairs(vv) do
                        -- Rank
                        if (vvv ~= nil) then
                            -- and (type(vvv) == "boolean") and (vvv) then
                            -- self:Debug("Value Value Key (type) ==> %s (%s)", vvk, type(vvv))                        
                            local spell = {}
                            local trainers = {}
                            local continue = true
                            -- local known = vvv.known
                            -- local met = false
                            
                            -- if (not includeKnown) then continue = not known end
                            if (not includeKnown) then continue = not vvv.known end
                          
                            if (continue) and (not teacher.spells[k]) then continue = false end
                            if (continue) and (not teacher.spells[k][vk]) then continue = false end
                            if (continue) and (not teacher.spells[k][vk][vvk]) then continue = false end
                            -- self:Debug("Category, Spell and Rank exist in account.")
                            
                            if (continue) then spell = self:Clone(teacher.spells[k][vk][vvk]) end
                            if (not spell) then continue = false end
                            -- spell.known = known
                            spell.known = vvv.known
                            spell.met = true
                            spell.criteria = {}
                            spell.criteria.level = {}
                            spell.criteria.level.met = true
                            spell.criteria.ability = {}
                            spell.criteria.ability.met = true
                            spell.criteria.skill = {}
                            spell.criteria.skill.met = true
                            
                            if (continue) and (student.level < spell.level) then continue = false end

                            if (continue) and (UnitLevel("player") < spell.level) then 
                                spell.met = false
                                spell.criteria.level.met = false
                            end
                            
                            if (continue) and (spell.ability ~= nil) and (not known) then
                                -- self:Debug("Check spell.ability")
                          
                                --[[
                                ["ability"] = {
                                    [1] = {
                                        ["skill"] = "Nature's Grasp (Rank 2)",
                                        ["req"] = 1,
                                    },
                                    [2] = {
                                        ["skill"] = "Entangling Roots (Rank 3)",
                                        ["req"] = 1,
                                    },
                                },  
                                --]]
                                local matches = {}
                                    
                                for a = 1, table.maxn(spell.ability) do
                                    if (not matches[a]) then matches[a] = {} end
                                    if (not matches[a].found) then
                                        local ability = spell.ability[a]["skill"]
                                        for s = 1, table.maxn(student.spells["player"]) do
                                            if (not matches[a].found) then
                                                local player = student.spells["player"][s].name .. " (" .. student.spells["player"][s].rank .. ")"
                                                if (ability == player) then matches[a].found = true end
                                            end
                                        end
                                            
                                        if (not matches[a].found) then
                                            --[[
                                            ["talents"] = {
                                                    ["Ferocity"] = {
                                                        ["maxrank"] = 5,
                                                        ["name"] = "Ferocity",
                                                        ["tier"] = 1,
                                                        ["column"] = 2,
                                                        ["icon"] = "Interface\\Icons\\Ability_Hunter_Pet_Hyena",
                                                        ["rank"] = 4,
                                                    },
                                            },                
                                            --]]
                                            for tk, tv in pairs(student.talents) do
                                                if (tv ~= nil) and (not matches[a].found) then
                                                    local talent = tv.name .. "(Rank " .. tv.rank .. ")"
                                                    if (ability == talent) then matches[a].found = true end
                                                end
                                            end
                                        end
                                    end
                                end

                                if (not matches) and (not School.db.profile.ignoreOverrideLevelTalent) then continue = false end
                                for m = 1, table.maxn(matches) do
                                    if (continue) and (not matches[m].found) then
                                        continue = false
                                    end
                                end
                                
                                if (not continue) then 
                                    spell.met = false
                                    spell.criteria.ability.met = false
                                    for c = 1, table.maxn(matches) do
                                        if (not matches[c].found) then
                                            table.insert(spell.criteria.ability, spell.ability[c])
                                        end
                                    end
                                end
                                
                                if (School.db.profile.ignoreOverrideLevelTalent) then continue = true end
                            end
                            
                            if (continue) and (spell.req ~= nil) and (not known) then
                                -- self:Debug("Check spell.req")
                                continue = false
                                if (spell.req.skill ~= nil) then
                                    if (student.skills[spell.req.skill] ~= nil) then
                                        -- student.skills[spell.req.skill].rank = 50
                                        if (not spell.req.rank) or (spell.req.rank == 0) then 
                                            continue = true
                                        else
                                            if (not School.db.profile.ignoreOverrideLevelSkill) then
                                                if (not School.db.profile.overrideLevelSkill) or (School.db.profile.overrideLevelSkill == 0) then
                                                    if (student.skills[spell.req.skill].rank >= spell.req.rank) then
                                                        continue = true
                                                    end
                                                else
                                                    if (School.db.profile.overrideLevelSkill >= spell.req.rank) then
                                                        continue = true
                                                    end
                                                end
                                            else
                                                -- local msg = string.format("You need to have %s (%s)", spell.req.skill, spell.req.rank)
                                                -- msg = msg .. "\n" .. string.format("You have %s (%s)", student.skills[spell.req.skill].name, student.skills[spell.req.skill].rank)
                                                -- self:LevelDebug(3, msg)
                                                if (student.skills[spell.req.skill].rank >= spell.req.rank) then
                                                    -- self:LevelDebug(3, "You qualify for %s", spell.req.skill)
                                                    continue = true
                                                end
                                            end
                                        end
                                        
                                        if (student.skills[spell.req.skill].rank < spell.req.rank) then 
                                            spell.met = false 
                                            spell.criteria.skill.met = false
                                            spell.criteria.skill.name = spell.req.skill
                                            spell.criteria.skill.rank = spell.req.rank
                                        end
                                    end
                                end
                            end
                            
                            if (continue) and (student.professions ~= nil) and (not known) then
                                -- self:Debug("Found student professions.")
                                
                                local primary = student.professions.primary
                                local secondary = student.professions.secondary
                                -- self:Debug("primary.count ==> %s", primary.count)
                                -- self:Debug("secondary.count ==> %s", secondary.count)                                    

                                local primaryContinue = true
                                local secondaryContinue = true
                                
                                -- Ugly hack start
                                -- student.professions.primary sometimes is NIL
                                if (not primary) or (not primary.count) then 
                                    primary = {}
                                    primary.count = 0
                                end
                                -- Ugly hack end
                                
                                -- Ugly hack start
                                -- student.profession.secondary sometimes is NIL
                                if (not secondary) or (not secondary.count) then
                                    secondary = {}
                                    secondary.count = 0
                                end
                                -- Ugly hack end
                                
                                -- self:Debug("spell.name ==> %s", spell.name)
                                if (primary.count == 2) then
                                    -- self:Debug("You already have 2 primary skills")
                                    if (apprentice.primary[spell.name]) then primaryContinue = false end
                                end
                                
                                if (secondary.count == 3) then
                                    -- self:Debug("You already have 3 secondary skills.")
                                    if (apprentice.secondary[spell.name]) then secondaryContinue = false end
                                end
                                
                                continue = primaryContinue and secondaryContinue
                            end
                            
                            if (continue) and (teacher.trainers ~= nil) then
                                spell.trainers = {}
                                local trainers = nil
                                
                                for tk, tv in pairs(teacher.trainers) do
                                    if (tv ~= nil) then
                                        if (tv.spells[spell.name] ~= nil) and (tv.spells[spell.name].rank) then
                                            if (spell.rank <= tv.spells[spell.name].rank) then
                                                local trainer = {}
                                                trainer.type = tv.type
                                                trainer.name = tv.name
                                                trainer.zone = tv.zone
                                                trainer.subzone = tv.subzone
                                                trainer.coords = tv.coords
                                                trainer.location = tv.zone
                                                if (trainer.subzone ~= nil) then trainer.location = trainer.subzone .. " (" .. trainer.location .. ")" end
                                                table.insert(spell.trainers, trainer)
                                            end
                                        end
                                    end
                                end
                            end

                            -- self:Debug("Known: %s", spell.known)
                            
                            if (continue) then
                                if (not school.spells) then school.spells = {} end
                                if (not school.spells[spell.category]) then school.spells[spell.category] = {} end
                                if (not school.spells[spell.category][spell.name]) then school.spells[spell.category][spell.name] = {} end
                                school.spells[spell.category][spell.name] = spell
                            end                            
                        end
                    end
                end
            end
        end
    end

    self.data.school = school
    
    --[[
    if (self.data.school ~= nil) then
        for k,v in pairs(self.data.school) do
            self:LevelDebug(3, "Key ==> %s", k)
        end
    end
    --]]
    
    return self.data.school
end

function School:GotoSchool()
    self:Debug("GotoSchool")
    
    self:BuildTeacher()
    self:BuildStudent()

    self:BuildSchool()

    self:TriggerEvent("School_Changed")
end

function School:FetchResults()
    -- self:Debug("FetchResults")
    local school = self:GetData()
    if (not school) then return end
        
    local show = {}
    local money = GetMoney()
    local spells = school["spells"]
    local trainers = school["trainers"]

    if (spells ~= nil) then
        for k,v in pairs(spells) do
            if (v ~= nil) then
                -- self:LevelDebug(3, "Key ==> %s", k)
                for vk, vv in pairs(v) do
                    -- self:LevelDebug(3, "Value Key ==> %s", vk) 
                    if (vv ~= nil) then
                        local spell = self:Clone(vv)
                        
                        --[[
                        if (trainers ~= nil) then
                            self:LevelDebug(3, "Found trainers.")
                            if (trainers[spell.name] ~= nil) then
                                self:LevelDebug(3, "Found trainers for " .. spell.name)
                                if (trainers[spell.name][spell.subtext] ~= nil) then
                                    self:LevelDebug(3, "Found trainers for " .. spell.name .. " Rank: " .. spell.subtext)
                                    if (not spell.trainers) then spell.trainers = {} end
                                    for tk, tv in pairs(trainers[spell.name][spell.subtext]) do
                                        if (tv ~= nil) then
                                            table.insert(spell.trainers, tv)
                                        end
                                    end
                                end
                            end
                        end
                        --]]
                        
                        --[[
                        if (trainers ~= nil) and (trainers[spell.name] ~= nil) and (trainers[spell.name][spell.subtext] ~= nil) then
                            if (not spell.trainers) then spell.trainers = {} end
                            for tk, tv in pairs(trainers[spell.name][spell.subtext]) do
                                if (tv ~= nil) then table.insert(spell.trainers, tv) end
                            end
                        end
                        --]]

                        if (not School.db.profile.ignoreCost) then
                            if (spell.cost.money ~= nil) and (money >= spell.cost.money) then table.insert(show, spell) end
                        else
                            table.insert(show, spell)
                        end
                    end
                end
            end
        end
    end

    local results = {}
    results.data = {}
    results.totals = {}
    results.totals.spells = {}
    results.totals.spells.known = 0
    results.totals.spells.met = 0
    results.totals.spells.all = 0
    results.totals.cost = {}
    results.totals.cost.known = 0
    results.totals.cost.met = 0
    results.totals.cost.all = 0
    
    results.count = 0
    results.totalcost = 0
    for k, v in pairs(show) do
        if (v ~= nil) then
            -- if (not results[k]) then results[k] = {} end
            -- results[k] = v
            if (not results.data[k]) then results.data[k] = {} end
            results.data[k] = v

            if (v.known) then 
                results.totals.spells.known = results.totals.spells.known + 1
                results.totals.cost.known = results.totals.cost.known + v.cost.money
            end

            if (v.met) and (not v.known) then 
                results.totals.spells.met = results.totals.spells.met + 1
                results.totals.cost.met = results.totals.cost.met + v.cost.money
            end
            
            results.totals.spells.all = results.totals.spells.all + 1
            results.totals.cost.all = results.totals.cost.all + v.cost.money
        end
    end

    --[[
    self:Debug("Totals\nSpells\nKnown: %s; Met: %s; All: %s\nCost\nKnown: %s; Met: %s; All: %s", 
        results.totals.spells.known,
        results.totals.spells.met,
        results.totals.spells.all,
        results.totals.cost.known,
        results.totals.cost.met,
        results.totals.cost.all
    )
    --]]

    results.count = results.totals.spells.all
    results.totalcost = results.totals.cost.all
    
    if (School:IsDebugging()) then School.db.char.FetchResults = results end
    return results
end

